from __future__ import annotations

import math
from typing import Literal

from PIL import Image

from labelle.lib.constants import Direction
from labelle.lib.env_config import is_dev_mode_no_margins
from labelle.lib.render_engines.render_context import RenderContext
from labelle.lib.render_engines.render_engine import (
    RenderEngine,
    RenderEngineException,
)


class BitmapTooBigError(RenderEngineException):
    def __init__(self, width_px, max_width_px):
        msg = f"width_px: {width_px}, max_width_px: {max_width_px}"
        super().__init__(msg)


class MarginsRenderEngine(RenderEngine):
    def __init__(
        self,
        render_engine: RenderEngine,
        mode: Literal["print", "preview"],
        justify: Direction = Direction.CENTER,
        visible_horizontal_margin_px: float = 0,
        labeler_margin_px: tuple[float, float] = (0, 0),
        max_width_px: float | None = None,
        min_width_px: float | None = 0,
    ):
        super().__init__()
        labeler_horizontal_margin_px, labeler_vertical_margin_px = labeler_margin_px
        assert visible_horizontal_margin_px >= 0
        assert labeler_horizontal_margin_px >= 0
        assert labeler_vertical_margin_px >= 0
        assert not max_width_px or max_width_px >= 0
        if min_width_px is None:
            min_width_px = 0
        assert min_width_px >= 0
        self.mode = mode
        self.justify = justify
        if is_dev_mode_no_margins():
            self.visible_horizontal_margin_px = 0.0
            self.labeler_horizontal_margin_px = 0.0
            self.min_width_px = 0.0
        else:
            self.visible_horizontal_margin_px = visible_horizontal_margin_px
            self.labeler_horizontal_margin_px = labeler_horizontal_margin_px
            self.min_width_px = min_width_px
        self.labeler_vertical_margin_px = labeler_vertical_margin_px
        self.max_width_px = max_width_px
        self.render_engine = render_engine

    def _calculate_visible_width(self, payload_width_px: int) -> float:
        minimal_label_width_px = (
            payload_width_px + self.visible_horizontal_margin_px * 2.0
        )
        if self.max_width_px is not None and minimal_label_width_px > self.max_width_px:
            raise BitmapTooBigError(minimal_label_width_px, self.max_width_px)

        if self.min_width_px > minimal_label_width_px:
            label_width_px = self.min_width_px
        else:
            label_width_px = minimal_label_width_px
        return label_width_px

    def render(self, _: RenderContext) -> Image.Image:
        raise RuntimeError("This should never be called")

    def render_with_meta(
        self, context: RenderContext
    ) -> tuple[Image.Image, dict[str, float]]:
        payload_bitmap = self.render_engine.render(context)
        payload_width_px = payload_bitmap.width
        label_width_px = self._calculate_visible_width(payload_width_px)
        padding_px = label_width_px - payload_width_px  # sum of margins from both sides

        if self.justify == Direction.LEFT:
            horizontal_offset_px = self.visible_horizontal_margin_px
        elif self.justify == Direction.CENTER:
            horizontal_offset_px = padding_px / 2
        elif self.justify == Direction.RIGHT:
            horizontal_offset_px = padding_px - self.visible_horizontal_margin_px
        assert horizontal_offset_px >= self.visible_horizontal_margin_px

        # In print mode:
        # ==============
        # There is a gap between the printer head and the cutter (for the sake of this
        # example, let us say it is DX pixels wide).
        # We assume the printing starts when the print head is in offset DX from the
        # label's edge (the label's edge begins just under the cutter).
        # After we print the payload, we need to offset the label DX pixels, in order
        # to move the edge of the printed payload past the cutter, otherwise the cutter
        # will cut inside the printed payload.
        # Subsequently, in order to achieve symmetry so that the final margin matches
        # the initial margin, we add another offset of DX pixels, so that the cut will
        # have that same margin from both sides of the payload edge. In summary, due to
        # the gap between the printer head and the cutter of DX pixels, printing an
        # initial margin of 0 pixels and a final margin of 2 * DX pixels leaves an
        # effective margin of DX pixels on both sides.
        #
        # There's also some vertical margin between printed area and the label
        # edge

        vertical_offset_px: float = 0
        if self.mode == "print":
            # print head is already in offset from label's edge under the cutter
            horizontal_offset_px -= self.labeler_horizontal_margin_px
            # no need to add vertical margins to bitmap
            bitmap_height = float(payload_bitmap.height)
        elif self.mode == "preview":
            # add vertical margins to bitmap
            bitmap_height = (
                float(payload_bitmap.height) + self.labeler_vertical_margin_px * 2.0
            )
            vertical_offset_px = self.labeler_vertical_margin_px

        bitmap = Image.new("1", (math.ceil(label_width_px), math.ceil(bitmap_height)))
        bitmap.paste(
            payload_bitmap, box=(round(horizontal_offset_px), round(vertical_offset_px))
        )
        meta = {
            "horizontal_offset_px": horizontal_offset_px,
            "vertical_offset_px": vertical_offset_px,
        }
        return bitmap, meta



# === LICENSE STATEMENT ===
# Copyright (c) 2011 Sebastian J. Bronner <waschtl@sbronner.com>
#
# Copying and distribution of this file, with or without modification, are
# permitted in any medium without royalty provided the copyright notice and
# this notice are preserved.
# === END LICENSE STATEMENT ===
from __future__ import annotations

import array
import logging
import math

import usb
from PIL import Image
from usb.core import NoBackendError, USBError

from labelle.lib.constants import ESC, SYN
from labelle.lib.devices.usb_device import UsbDevice, UsbDeviceError
from labelle.lib.utils import mm_to_px

LOG = logging.getLogger(__name__)
POSSIBLE_USB_ERRORS = (UsbDeviceError, NoBackendError, USBError)


class DymoLabelerDetectError(Exception):
    def __init__(self, error: str):
        msg = f"Detection error: {error}"
        super().__init__(msg)


class DymoLabelerPrintError(Exception):
    def __init__(self, error: str):
        msg = f"Print error: {error}"
        super().__init__(msg)


class DymoLabelerFunctions:
    """Create and work with a Dymo LabelManager PnP object.

    This class contains both mid-level and high-level functions. In general,
    the high-level functions should be used. However, special purpose usage
    may require the mid-level functions. That is why they are provided.
    However, they should be well understood before use. Look at the
    high-level functions for help. Each function is marked in its docstring
    with 'HLF' or 'MLF' in parentheses.

    A partial reference of the protocol is the Technical Reference for the
    LabelWriter 450:
    <https://download.dymo.com/dymo/technical-data-sheets/LW%20450%20Series%20Technical%20Reference.pdf>
    """

    # Max number of print lines to send before waiting for a response. This helps
    # to avoid timeouts due to differences between data transfer and
    # printer speeds. I added this because I kept getting "IOError: [Errno
    # 110] Connection timed out" with long labels. Using dev.default_timeout
    # (1000) and the transfer speeds available in the descriptors somewhere, a
    # sensible timeout can also be calculated dynamically.
    _synwait: int | None
    _bytesPerLine: int | None
    _devout: usb.core.Endpoint
    _devin: usb.core.Endpoint

    def __init__(
        self,
        devout: usb.core.Endpoint,
        devin: usb.core.Endpoint,
        synwait: int | None = None,
    ):
        """Initialize the LabelManager object (HLF)."""
        self._cmd: list[int] = []
        self._response = False
        self._bytesPerLine = None
        self._dotTab = 0
        self._maxLines = 200
        self._devout = devout
        self._devin = devin
        self._synwait = synwait

    @classmethod
    def _max_bytes_per_line(cls, tape_size_mm: int) -> int:
        return int(8 * tape_size_mm / 12)

    @classmethod
    def height_px(cls, tape_size_mm: int):
        return cls._max_bytes_per_line(tape_size_mm) * 8

    def _send_command(self):
        """Send the already built command to the LabelManager (MLF)."""
        if len(self._cmd) == 0:
            return None

        while len(self._cmd) > 0:
            if self._synwait is None:
                cmd_to_send = self._cmd
                cmd_rest = []
            else:
                # Send a status request
                cmdBin = array.array("B", [ESC, ord("A")])
                cmdBin.tofile(self._devout)
                rspBin = self._devin.read(512)
                _ = array.array("B", rspBin).tolist()
                # Ok, we got a response. Now we can send a chunk of data

                # Compute a chunk with at most synwait SYN characters
                synCount = 0  # Number of SYN characters encountered in iteration
                pos = -1  # Index of last SYN character encountered in iteration
                while synCount < self._synwait:
                    try:
                        # Increment pos to the index of the next SYN character
                        pos += self._cmd[pos + 1 :].index(SYN) + 1
                        synCount += 1
                    except ValueError:
                        # No more SYN characters in cmd
                        pos = len(self._cmd)
                        break
                cmd_to_send = self._cmd[:pos]
                cmd_rest = self._cmd[pos:]
                LOG.debug(f"Sending chunk of {len(cmd_to_send)} bytes")

            # Remove the computed chunk from the command to be processed
            self._cmd = cmd_rest

            # Send the chunk
            cmdBin = array.array("B", cmd_to_send)
            cmdBin.tofile(self._devout)

        self._cmd = []  # This looks redundant.
        if not self._response:
            return None
        self._response = False
        responseBin = self._devin.read(512)
        response = array.array("B", responseBin).tolist()
        return response

    def _reset_command(self) -> None:
        """Remove a partially built command (MLF)."""
        self._cmd = []
        self._response = False

    def _build_command(self, cmd):
        """Add the next instruction to the command (MLF)."""
        self._cmd += cmd

    def _status_request(self) -> None:
        """Set instruction to get the device's status (MLF)."""
        cmd = [ESC, ord("A")]
        self._build_command(cmd)
        self._response = True

    def _dot_tab(self, value, tape_size_mm: int) -> None:
        """Set the bias text height, in bytes (MLF)."""
        if value < 0 or value > self._max_bytes_per_line(tape_size_mm):
            raise ValueError
        cmd = [ESC, ord("B"), value]
        self._build_command(cmd)
        self._dotTab = value
        self._bytesPerLine = None

    def _tape_color(self, value):
        """Set the tape color (MLF)."""
        if value < 0:
            raise ValueError
        cmd = [ESC, ord("C"), value]
        self._build_command(cmd)

    def _bytes_per_line(self, value: int):
        """Set the number of bytes sent in the following lines (MLF)."""
        if value == self._bytesPerLine:
            return
        cmd = [ESC, ord("D"), value]
        self._build_command(cmd)
        self._bytesPerLine = value

    def _cut(self) -> None:
        """Set instruction to trigger cutting of the tape (MLF)."""
        cmd = [ESC, ord("E")]
        self._build_command(cmd)

    def _line(self, value) -> None:
        """Set next printed line (MLF)."""
        self._bytes_per_line(len(value))
        cmd = [SYN, *value]
        self._build_command(cmd)

    def _chain_mark(self, tape_size_mm: int) -> None:
        """Set Chain Mark (MLF)."""
        self._dot_tab(0, tape_size_mm)
        self._bytes_per_line(self._max_bytes_per_line(tape_size_mm))
        self._line([0x99] * self._max_bytes_per_line(tape_size_mm))

    def _skip_lines(self, value) -> None:
        """Set number of lines of white to print (MLF)."""
        if value <= 0:
            raise ValueError
        self._bytes_per_line(0)
        cmd = [SYN] * value
        self._build_command(cmd)

    def _init_label(self) -> None:
        """Set the label initialization sequence (MLF).

        This was in the original dymoprint by S. Bronner but was never invoked.
        (There was a self.initLabel without parentheses.)
        I see no mention of it in the technical reference, so this seems to be
        dead code.
        """
        cmd = [0x00] * 8
        self._build_command(cmd)

    def _get_status(self):
        """Ask for and return the device's status (HLF)."""
        self._status_request()
        return self._send_command()

    def print_label(self, lines: list[list[int]]):
        """Print the label described by lines.

        Automatically split the label if it's larger than maxLines.
        """
        while len(lines) > self._maxLines + 1:
            self._raw_print_label(lines[0 : self._maxLines])
            del lines[0 : self._maxLines]
        self._raw_print_label(lines)

    def _raw_print_label(self, lines: list[list[int]]):
        """Print the label described by lines (HLF)."""
        # Here used to be a matrix optimization code that caused problems in issue #87
        self._tape_color(0)
        for line in lines:
            self._line(line)
        self._status_request()
        status = self._get_status()
        LOG.debug(f"Post-send response: {status}")


class DymoLabeler:
    _device: UsbDevice | None
    tape_size_mm: int

    LABELER_DISTANCE_BETWEEN_PRINT_HEAD_AND_CUTTER_MM = 8.1
    LABELER_PRINT_HEAD_HEIGHT_MM = 8.2
    SUPPORTED_TAPE_SIZES_MM = (19, 12, 9, 6)
    DEFAULT_TAPE_SIZE_MM = 12

    def __init__(
        self,
        tape_size_mm: int | None = None,
        device: UsbDevice | None = None,
    ):
        if tape_size_mm is None:
            tape_size_mm = self.DEFAULT_TAPE_SIZE_MM
        if tape_size_mm not in self.SUPPORTED_TAPE_SIZES_MM:
            raise ValueError(
                f"Unsupported tape size {tape_size_mm}mm. "
                f"Supported sizes: {self.SUPPORTED_TAPE_SIZES_MM}"
            )
        self.tape_size_mm = tape_size_mm
        self._device = device

    @property
    def height_px(self):
        return DymoLabelerFunctions.height_px(self.tape_size_mm)

    @property
    def _functions(self) -> DymoLabelerFunctions:
        assert self._device is not None
        return DymoLabelerFunctions(
            devout=self._device.devout,
            devin=self._device.devin,
            synwait=64,
        )

    @property
    def minimum_horizontal_margin_mm(self):
        return self.LABELER_DISTANCE_BETWEEN_PRINT_HEAD_AND_CUTTER_MM

    @property
    def labeler_margin_px(self) -> tuple[float, float]:
        vertical_margin_mm = max(
            0, (self.tape_size_mm - self.LABELER_PRINT_HEAD_HEIGHT_MM) / 2
        )

        return (
            mm_to_px(self.minimum_horizontal_margin_mm),
            mm_to_px(vertical_margin_mm),
        )

    @property
    def device(self) -> UsbDevice | None:
        return self._device

    @device.setter
    def device(self, device: UsbDevice | None):
        try:
            if device:
                device.setup()
        except UsbDeviceError as e:
            device = None
            LOG.error(e)
        self._device = device

    @property
    def is_ready(self) -> bool:
        return self.device is not None

    def print(
        self,
        bitmap: Image.Image,
    ) -> None:
        """Print a label bitmap to the detected printer.

        The label bitmap is a PIL image in 1-bit format (mode=1), and pixels with value
        equal to 1 are burned.
        """
        # Convert the image to the proper matrix for the dymo labeler object so that
        # rows span the width of the label, and the first row corresponds to the left
        # edge of the label.
        rotated_bitmap = bitmap.transpose(Image.Transpose.ROTATE_270)

        # Convert the image to raw bytes. Pixels along rows are chunked into groups of
        # 8 pixels, and subsequent rows are concatenated.
        stream: bytes = rotated_bitmap.tobytes()

        # Regather the bytes into rows
        stream_row_length = math.ceil(bitmap.height / 8)
        if len(stream) // stream_row_length != bitmap.width:
            raise RuntimeError(
                "An internal problem was encountered while processing the label bitmap!"
            )
        label_rows: list[bytes] = [
            stream[i : i + stream_row_length]
            for i in range(0, len(stream), stream_row_length)
        ]

        # Convert bytes into ints
        label_matrix: list[list[int]] = [
            array.array("B", label_row).tolist() for label_row in label_rows
        ]

        try:
            LOG.debug("Printing label..")
            self._functions.print_label(label_matrix)
            LOG.debug("Done printing.")
            if self._device is not None:
                self._device.dispose()
            LOG.debug("Cleaned up.")
        except POSSIBLE_USB_ERRORS as e:
            raise DymoLabelerPrintError(str(e)) from e



from labelle.lib.render_engines.barcode import BarcodeRenderEngine, BarcodeRenderError
from labelle.lib.render_engines.barcode_with_text import BarcodeWithTextRenderEngine
from labelle.lib.render_engines.empty import EmptyRenderEngine
from labelle.lib.render_engines.exceptions import NoContentError
from labelle.lib.render_engines.horizontally_combined import (
    HorizontallyCombinedRenderEngine,
)
from labelle.lib.render_engines.margins import MarginsRenderEngine
from labelle.lib.render_engines.picture import (
    PicturePathDoesNotExist,
    PictureRenderEngine,
    UnidentifiedImageFileError,
)
from labelle.lib.render_engines.print_payload import PrintPayloadRenderEngine
from labelle.lib.render_engines.print_preview import PrintPreviewRenderEngine
from labelle.lib.render_engines.qr import QrRenderEngine, QrTooBigError
from labelle.lib.render_engines.render_context import RenderContext
from labelle.lib.render_engines.render_engine import RenderEngine
from labelle.lib.render_engines.sample_pattern import SamplePatternRenderEngine
from labelle.lib.render_engines.text import TextRenderEngine

__all__ = [
    "BarcodeRenderEngine",
    "BarcodeRenderError",
    "BarcodeWithTextRenderEngine",
    "EmptyRenderEngine",
    "HorizontallyCombinedRenderEngine",
    "MarginsRenderEngine",
    "NoContentError",
    "PicturePathDoesNotExist",
    "PictureRenderEngine",
    "PrintPayloadRenderEngine",
    "PrintPreviewRenderEngine",
    "QrRenderEngine",
    "QrTooBigError",
    "RenderContext",
    "RenderEngine",
    "SamplePatternRenderEngine",
    "TextRenderEngine",
    "UnidentifiedImageFileError",
]



from __future__ import annotations

import math

from PIL import Image, ImageDraw, ImageFont

from labelle.lib.font_config import get_font_path
from labelle.lib.render_engines import RenderContext, RenderEngine
from labelle.lib.render_engines.horizontally_combined import (
    HorizontallyCombinedRenderEngine,
)
from labelle.lib.utils import draw_image

FONT_SIZE_PX = 12


class SamplePatternRenderEngine(HorizontallyCombinedRenderEngine):
    def __init__(self, height: int = 100):
        four_horizontal_lines_top_and_bottom = (
            _make_staggered_horizontal_lines_top_bottom(
                num_lines=4, width=40, height=height
            )
        )
        vertical_lines = _make_vertical_lines(num_lines=5, height=height)
        fine_checkerboard_pattern = _make_fine_checkerboard_pattern(
            width=12, height=height
        )
        solid_black = _make_solid_black(width=12, height=height)
        dyadic_checkerboard_pattern = _make_dyadic_checkerboard_pattern(height=height)

        bitmaps = [
            four_horizontal_lines_top_and_bottom,
            vertical_lines,
            fine_checkerboard_pattern,
            solid_black,
            dyadic_checkerboard_pattern,
            four_horizontal_lines_top_and_bottom,
        ]
        render_engines = [_ImageRenderEngine(bitmap) for bitmap in bitmaps]
        super().__init__(render_engines=render_engines, padding=0)


class _ImageRenderEngine(RenderEngine):
    def __init__(self, image: Image.Image):
        super().__init__()
        self.image = image

    def render(self, _: RenderContext) -> Image.Image:
        return self.image


def _make_vertical_lines(*, num_lines: int, height: int) -> Image.Image:
    width = 2 * num_lines - 1
    image = Image.new("1", (width, height))
    for x in range(0, width, 2):
        for y in range(height):
            image.putpixel((x, y), 1)
    return image


def _make_staggered_horizontal_lines_top_bottom(
    *, num_lines: int, width: int, height: int
) -> Image.Image:
    image = Image.new("1", (40, height))
    # top
    for y0 in range(0, 2 * num_lines, 2):
        for x in range(width):
            y = y0 if x < width / 2 else y0 + 1
            image.putpixel((x, y), 1)
    # bottom
    for y0 in range(height - 2 * num_lines, height, 2):
        for x in range(40):
            y = y0 + 1 if x < width / 2 else y0
            image.putpixel((x, y), 1)
    # draw height
    font_path = get_font_path(style="regular")
    font = ImageFont.truetype(str(font_path), FONT_SIZE_PX)
    with draw_image(image) as draw:
        assert isinstance(draw, ImageDraw.ImageDraw)
        text = f"h={height}"
        _left, _top, _right, text_height = font.getbbox(text)
        y = (height - text_height) // 2
        X_OFFSET = 3
        draw.text((X_OFFSET, y), text, font=font, fill=1)
    return image


def _make_fine_checkerboard_pattern(*, width: int, height: int) -> Image.Image:
    image = Image.new("1", (width, height))
    for x in range(0, width):
        for y in range(0, height):
            if (x + y) % 2 == 0:
                image.putpixel((x, y), 1)
    return image


def _make_dyadic_checkerboard_pattern(*, height: int) -> Image.Image:
    font_path = get_font_path(style="regular")
    font = ImageFont.truetype(str(font_path), FONT_SIZE_PX)
    _left, _top, _right, font_height = font.getbbox("0123456789")
    MARGIN_BELOW = 3  # One pixel above plus two pixels below the text.

    # Font block size is the first power of two greater than the required height.
    log_font_block_size = math.ceil(math.log2(font_height + MARGIN_BELOW))
    font_block_size = 2**log_font_block_size

    required_text_width = _get_required_text_width(
        font=font, height=height, log_font_block_size=log_font_block_size
    )
    text_x_offset = log_font_block_size * font_block_size
    image_width = text_x_offset + required_text_width + 2

    image = Image.new("1", (image_width, height))

    # Draw text
    for yc in range(font_block_size, height + 1, font_block_size):
        text = str(yc)
        y = height - yc - 1
        with draw_image(image) as draw:
            assert isinstance(draw, ImageDraw.ImageDraw)
            draw.text((text_x_offset + 1, y + 2), text, font=font, fill=1)

    # Draw checkerboard pattern
    for yc in range(0, height):
        y = height - yc - 1
        for x in range(image_width):
            x0 = min(x // font_block_size, log_font_block_size)
            # Get the x0-th bit of yc
            if (yc >> x0) & 1 == 0:
                if image.getpixel((x, y)) == 1:
                    image.putpixel((x, y), 0)
                else:
                    image.putpixel((x, y), 1)
    return image


def _get_required_text_width(
    *, font: ImageFont.FreeTypeFont, height: int, log_font_block_size: int
) -> int:
    font_block_size = 2**log_font_block_size
    required_text_width = 0
    for yc in range(font_block_size, height + 1, font_block_size):
        text = str(yc)
        _left, _top, text_width, _bottom = font.getbbox(text)
        required_text_width = max(text_width, required_text_width)
    return required_text_width


def _make_solid_black(*, width: int, height: int) -> Image.Image:
    return Image.new("1", (width, height), color=1)



from PIL.Image import Image
from PIL.ImageOps import pad

UH = "▀"
LH = "▄"
FB = "█"
NB = "\xa0"

assert UH == "\N{UPPER HALF BLOCK}"
assert LH == "\N{LOWER HALF BLOCK}"
assert FB == "\N{FULL BLOCK}"
assert NB == "\N{NO-BREAK SPACE}"

BLACK_PIXEL = (255, 255, 255, 255)
WHITE_PIXEL = (0, 0, 0, 255)
# when the image height is odd, the preview image height is extended by 1 pixel
# this padding result in the added line being composed of transparent pixels.
TRANSPARENT_PIXEL = (0, 0, 0, 0)

dict_unicode = {
    (BLACK_PIXEL, BLACK_PIXEL): FB,
    (WHITE_PIXEL, BLACK_PIXEL): LH,
    (BLACK_PIXEL, WHITE_PIXEL): UH,
    (WHITE_PIXEL, WHITE_PIXEL): NB,
    (BLACK_PIXEL, TRANSPARENT_PIXEL): UH,
    (WHITE_PIXEL, TRANSPARENT_PIXEL): NB,
}

dict_unicode_inverted = {
    (BLACK_PIXEL, BLACK_PIXEL): NB,
    (WHITE_PIXEL, BLACK_PIXEL): UH,
    (BLACK_PIXEL, WHITE_PIXEL): LH,
    (WHITE_PIXEL, WHITE_PIXEL): FB,
    (BLACK_PIXEL, TRANSPARENT_PIXEL): NB,
    (WHITE_PIXEL, TRANSPARENT_PIXEL): UH,
}


def image_to_unicode(im: Image, invert: bool = False) -> str:
    char_for = dict_unicode_inverted if invert else dict_unicode
    width = im.width
    height = im.height + (im.height % 2)
    padded_im = pad(image=im, size=(width, height))
    a = padded_im.load()
    output_rows = []
    for r in range(0, height, 2):
        char_list = [char_for[(a[c, r], a[c, r + 1])] for c in range(width)]
        row = "".join(char_list)
        output_rows.append(row)
    output_str = "\n".join(output_rows)
    return output_str



from pathlib import Path

import PIL.ImageOps
import pytest

from labelle.lib.constants import DEFAULT_BARCODE_TYPE, BarcodeType, Direction
from labelle.lib.render_engines import (
    BarcodeRenderEngine,
    BarcodeRenderError,
    BarcodeWithTextRenderEngine,
    EmptyRenderEngine,
    HorizontallyCombinedRenderEngine,
    NoContentError,
    PicturePathDoesNotExist,
    PictureRenderEngine,
    QrRenderEngine,
    QrTooBigError,
    RenderContext,
    SamplePatternRenderEngine,
    TextRenderEngine,
    UnidentifiedImageFileError,
)

RENDER_CONTEXT = RenderContext(height_px=100)
TESTS_DIR = Path(__file__).parent
FONT_FILE_NAME = "src/labelle/resources/fonts/Carlito-Regular.ttf"
EXPECTED_RENDERS_DIR = TESTS_DIR.joinpath("expected_renders")
OUTPUT_RENDER = TESTS_DIR.joinpath("output.png")
FONT_SIZE_RATIOS = [x / 10 for x in range(2, 11, 2)]


def verify_image(request, image_diff, image):
    filename = Path(request.node.name.replace(".", "_")).with_suffix(".png")
    actual = TESTS_DIR.joinpath(filename)
    inverted = PIL.ImageOps.invert(image.convert("RGB"))
    inverted.save(actual)
    expected = EXPECTED_RENDERS_DIR.joinpath(filename)
    # Note: threshold should be omitted in the future
    # (CI bitmaps are different than development machine)
    image_diff(expected, actual, threshold=0.15)
    actual.unlink()


###############################
# BarcodeWithTextRenderEngine #
###############################


def test_barcode_with_text_render_engine(request, image_diff):
    render_engine = BarcodeWithTextRenderEngine(
        content="hello, world!",
        font_file_name=FONT_FILE_NAME,
    )
    image = render_engine.render(RENDER_CONTEXT)
    verify_image(request, image_diff, image)


@pytest.mark.parametrize("align", Direction)
def test_barcode_with_text_render_engine_alignment(request, image_diff, align):
    render_engine = BarcodeWithTextRenderEngine(
        content="hello, world!",
        font_file_name=FONT_FILE_NAME,
        align=align,
    )
    image = render_engine.render(RENDER_CONTEXT)
    verify_image(request, image_diff, image)


@pytest.mark.parametrize("font_size_ratio", [x / 10 for x in range(2, 11, 2)])
def test_barcode_with_text_render_engine_font_size_ratio(
    request, image_diff, font_size_ratio
):
    render_engine = BarcodeWithTextRenderEngine(
        content="hello, world!",
        font_file_name=FONT_FILE_NAME,
        font_size_ratio=font_size_ratio,
    )
    image = render_engine.render(RENDER_CONTEXT)
    verify_image(request, image_diff, image)


#######################
# BarcodeRenderEngine #
#######################


def test_barcode_render_engine(request, image_diff):
    render_engine = BarcodeRenderEngine(
        content="hello, world!",
    )
    image = render_engine.render(RENDER_CONTEXT)
    verify_image(request, image_diff, image)


@pytest.mark.parametrize(
    "barcode_type,content",
    [
        (BarcodeType.EAN, "123456789012"),
        (BarcodeType.UPC, "12345678901"),
        (BarcodeType.CODE39, "123"),
        (DEFAULT_BARCODE_TYPE, " "),
        # Empty string in default type should be replaced with a space
        (DEFAULT_BARCODE_TYPE, ""),
    ],
)
def test_barcode_render_engine_barcode_type(request, image_diff, barcode_type, content):
    render_engine = BarcodeRenderEngine(content=content, barcode_type=barcode_type)
    image = render_engine.render(RENDER_CONTEXT)
    verify_image(request, image_diff, image)


def test_barcode_render_engine_default_barcode_type():
    """Test default type handling.

    Test our mechanism of replacing an empty string with a space for the default
    barcode type, to prevent an error.
    """
    images = [
        BarcodeRenderEngine(content=content, barcode_type=DEFAULT_BARCODE_TYPE).render(
            RENDER_CONTEXT
        )
        for content in ["", " "]
    ]
    assert images[0].tobytes() == images[1].tobytes()


def test_barcode_render_engine_internal_error():
    render_engine = BarcodeRenderEngine(
        content="No alphabet allowed", barcode_type=BarcodeType.EAN
    )
    with pytest.raises(BarcodeRenderError) as exc_info:
        render_engine.render(RENDER_CONTEXT)
    assert str(exc_info.value).startswith(
        "Barcode render error: IllegalCharacterError('EAN code can only contain numbers"
    )


#####################
# EmptyRenderEngine #
#####################


@pytest.mark.parametrize("width_px", [1, 10, 100])
def test_empty_render_engine(request, image_diff, width_px):
    render_engine = EmptyRenderEngine(
        width_px=width_px,
    )
    image = render_engine.render(RENDER_CONTEXT)
    verify_image(request, image_diff, image)


####################################
# HorizontallyCombinedRenderEngine #
####################################


def test_horizontally_combined_render_engine_single(request, image_diff):
    inner_render_engine = TextRenderEngine(
        text_lines=["Hello, World!"],
        font_file_name=FONT_FILE_NAME,
    )
    render_engine = HorizontallyCombinedRenderEngine(
        render_engines=[inner_render_engine]
    )
    image = render_engine.render(RENDER_CONTEXT)
    verify_image(request, image_diff, image)


def test_horizontally_combined_render_engine_multiple(request, image_diff):
    inner_render_engines = [
        TextRenderEngine(
            text_lines=[f"Render #{i}"],
            font_file_name=FONT_FILE_NAME,
        )
        for i in range(1, 4)
    ]
    render_engine = HorizontallyCombinedRenderEngine(
        render_engines=inner_render_engines
    )
    image = render_engine.render(RENDER_CONTEXT)
    verify_image(request, image_diff, image)


def test_horizontally_combined_render_engine_empty(request, image_diff):
    render_engine = HorizontallyCombinedRenderEngine(render_engines=[])
    image = render_engine.render(RENDER_CONTEXT)
    verify_image(request, image_diff, image)


#######################
# PictureRenderEngine #
#######################


def test_picture_render_engine(request, image_diff):
    render_engine = PictureRenderEngine(picture_path="labelle.png")
    image = render_engine.render(RENDER_CONTEXT)
    verify_image(request, image_diff, image)


def test_picture_render_engine_bad_path():
    with pytest.raises(PicturePathDoesNotExist) as exc_info:
        PictureRenderEngine(picture_path="non_existent.png")
    assert str(exc_info.value) == "Picture path does not exist: non_existent.png"


def test_picture_render_engine_empty_path():
    with pytest.raises(NoContentError):
        PictureRenderEngine(picture_path="")


def test_picture_render_engine_bad_image_file():
    render_engine = PictureRenderEngine(picture_path="README.md")
    with pytest.raises(UnidentifiedImageFileError) as exc_info:
        render_engine.render(RENDER_CONTEXT)
    assert str(exc_info.value).startswith("cannot identify image file")


##################
# QrRenderEngine #
##################


def test_qr_render_engine(request, image_diff):
    render_engine = QrRenderEngine(content="Hello, World!")
    image = render_engine.render(RENDER_CONTEXT)
    verify_image(request, image_diff, image)


def test_qr_render_engine_no_content():
    with pytest.raises(NoContentError):
        QrRenderEngine(content="")


def test_qr_render_engine_too_big():
    render_engine = QrRenderEngine(content="Hello, World!" * 100)
    with pytest.raises(QrTooBigError) as exc_info:
        render_engine.render(RENDER_CONTEXT)
    assert str(exc_info.value) == "Too much information to store in the QR code"


#############################
# SamplePatternRenderEngine #
#############################


@pytest.mark.parametrize("height", [15, 64, 65, 100, 256])
def test_sample_pattern_render_engine(request, image_diff, height):
    render_engine = SamplePatternRenderEngine(height=height)
    image = render_engine.render(RENDER_CONTEXT)
    verify_image(request, image_diff, image)


####################
# TextRenderEngine #
####################


def test_text_render_engine_single_line(request, image_diff):
    render_engine = TextRenderEngine(
        text_lines=["Hello, World!"],
        font_file_name=FONT_FILE_NAME,
    )
    image = render_engine.render(RENDER_CONTEXT)
    verify_image(request, image_diff, image)


def test_text_render_engine_with_frame(request, image_diff):
    render_engine = TextRenderEngine(
        text_lines=["Hello, World!"], font_file_name=FONT_FILE_NAME, frame_width_px=5
    )
    image = render_engine.render(RENDER_CONTEXT)
    verify_image(request, image_diff, image)


def test_text_render_engine_with_multiple_lines(request, image_diff):
    render_engine = TextRenderEngine(
        text_lines=["Hello,", "World!"],
        font_file_name=FONT_FILE_NAME,
    )
    image = render_engine.render(RENDER_CONTEXT)
    verify_image(request, image_diff, image)


@pytest.mark.parametrize("align", Direction)
def test_text_render_engine_alignment(request, image_diff, align):
    render_engine = TextRenderEngine(
        text_lines=["Hi,", "World!"],
        font_file_name=FONT_FILE_NAME,
        align=align,
    )
    image = render_engine.render(RENDER_CONTEXT)
    verify_image(request, image_diff, image)


@pytest.mark.parametrize("font_size_ratio", FONT_SIZE_RATIOS)
def test_text_render_engine_font_size_ratio(request, image_diff, font_size_ratio):
    render_engine = TextRenderEngine(
        text_lines=["Hello, World!"],
        font_file_name=FONT_FILE_NAME,
        font_size_ratio=font_size_ratio,
    )
    image = render_engine.render(RENDER_CONTEXT)
    verify_image(request, image_diff, image)


def test_text_render_engine_empty_line(request, image_diff):
    render_engine = TextRenderEngine(
        text_lines=[],
        font_file_name=FONT_FILE_NAME,
    )
    image = render_engine.render(RENDER_CONTEXT)
    verify_image(request, image_diff, image)


def test_text_render_engine_empty_lines(request, image_diff):
    render_engine = TextRenderEngine(
        text_lines=[],
        font_file_name=FONT_FILE_NAME,
    )
    image = render_engine.render(RENDER_CONTEXT)
    verify_image(request, image_diff, image)



from __future__ import annotations

from PIL import Image

from labelle.lib.constants import Direction
from labelle.lib.render_engines.margins import MarginsRenderEngine
from labelle.lib.render_engines.render_context import RenderContext
from labelle.lib.render_engines.render_engine import RenderEngine


class PrintPayloadRenderEngine(RenderEngine):
    def __init__(
        self,
        render_engine: RenderEngine,
        justify: Direction = Direction.CENTER,
        visible_horizontal_margin_px: float = 0,
        labeler_margin_px: tuple[float, float] = (0, 0),
        max_width_px: float | None = None,
        min_width_px: float | None = 0,
    ):
        super().__init__()
        self.render_engine = MarginsRenderEngine(
            render_engine=render_engine,
            mode="print",
            justify=justify,
            visible_horizontal_margin_px=visible_horizontal_margin_px,
            labeler_margin_px=labeler_margin_px,
            max_width_px=max_width_px,
            min_width_px=min_width_px,
        )

    def render(self, _: RenderContext) -> Image.Image:
        raise RuntimeError("This should never be called")

    def render_with_meta(
        self, context: RenderContext
    ) -> tuple[Image.Image, dict[str, float]]:
        return self.render_engine.render_with_meta(context)



from __future__ import annotations

from darkdetect import isDark
from PIL import Image, ImageColor, ImageDraw, ImageOps

from labelle.lib.constants import Direction
from labelle.lib.render_engines.margins import MarginsRenderEngine
from labelle.lib.render_engines.render_context import RenderContext
from labelle.lib.render_engines.render_engine import RenderEngine
from labelle.lib.utils import px_to_mm


class PrintPreviewRenderEngine(RenderEngine):
    X_MARGIN_PX = 80
    Y_MARGIN_PX = 30
    DX = X_MARGIN_PX * 0.3
    DY = Y_MARGIN_PX * 0.3

    def __init__(
        self,
        render_engine: RenderEngine,
        justify: Direction = Direction.CENTER,
        visible_horizontal_margin_px: float = 0,
        labeler_margin_px: tuple[float, float] = (0, 0),
        max_width_px: float | None = None,
        min_width_px: float = 0,
    ):
        super().__init__()
        self.render_engine = MarginsRenderEngine(
            render_engine=render_engine,
            mode="preview",
            justify=justify,
            visible_horizontal_margin_px=visible_horizontal_margin_px,
            labeler_margin_px=labeler_margin_px,
            max_width_px=max_width_px,
            min_width_px=min_width_px,
        )

    @staticmethod
    def _get_margin_color():
        return "red" if isDark() else "gray"

    @staticmethod
    def _get_mark_color():
        return "yellow" if isDark() else "red"

    @staticmethod
    def _get_text_color():
        return "white" if isDark() else "blue"

    def _get_label_bitmap(self, context: RenderContext):
        render_bitmap, meta = self.render_engine.render_with_meta(context)
        bitmap = ImageOps.invert(render_bitmap.convert("L")).convert("RGBA")
        pixel_map = {
            "black": context.foreground_color,
            "white": context.background_color,
        }
        pixel_color_map = {
            ImageColor.getcolor(k, "RGBA"): ImageColor.getcolor(v, "RGBA")
            for k, v in pixel_map.items()
        }
        pixdata = bitmap.load()
        width, height = bitmap.size
        for x in range(0, width):
            for y in range(0, height):
                pixdata[x, y] = pixel_color_map[pixdata[x, y]]
        return bitmap, meta

    def _show_margins(self, label_bitmap, preview_bitmap, meta, context):
        draw = ImageDraw.Draw(preview_bitmap)
        x_margin = meta["horizontal_offset_px"]
        y_margin = meta["vertical_offset_px"]
        preview_width, preview_height = preview_bitmap.size
        label_width, label_height = label_bitmap.size

        preview_width_mark_y = preview_height - self.Y_MARGIN_PX + self.DY
        label_width_mark_y = preview_height - self.DY
        preview_width_mark_x = self.X_MARGIN_PX - self.DX
        label_width_mark_x = self.DX
        margin_color = self._get_margin_color()
        mark_color = self._get_mark_color()
        text_color = self._get_text_color()

        # left vertical margin
        draw.line(
            xy=(
                self.X_MARGIN_PX + x_margin,
                0,
                self.X_MARGIN_PX + x_margin,
                preview_width_mark_y,
            ),
            fill=margin_color,
        )
        # right vertical margin
        draw.line(
            xy=(
                self.X_MARGIN_PX + label_width - x_margin,
                0,
                self.X_MARGIN_PX + label_width - x_margin,
                preview_width_mark_y,
            ),
            fill=margin_color,
        )
        # top horizontal margin
        draw.line(
            xy=(
                preview_width_mark_x,
                self.DY + y_margin,
                preview_width,
                self.DY + y_margin,
            ),
            fill=margin_color,
        )
        # bottom horizontal margin
        draw.line(
            xy=(
                preview_width_mark_x,
                self.DY + label_height - y_margin,
                preview_width,
                self.DY + label_height - y_margin,
            ),
            fill=margin_color,
        )
        # horizontal line for payload width
        draw.line(
            xy=(
                self.X_MARGIN_PX + x_margin,
                preview_width_mark_y,
                self.X_MARGIN_PX + label_width - x_margin,
                preview_width_mark_y,
            ),
            fill=mark_color,
        )
        # horizontal line for label width
        draw.line(
            xy=(
                self.X_MARGIN_PX,
                label_width_mark_y,
                self.X_MARGIN_PX + label_width,
                label_width_mark_y,
            ),
            fill=mark_color,
        )
        # vertical line for payload height
        draw.line(
            xy=(
                preview_width_mark_x,
                self.DY + y_margin,
                preview_width_mark_x,
                self.DY + label_height - y_margin,
            ),
            fill=mark_color,
        )
        # vertical line for label height
        draw.line(
            xy=(
                label_width_mark_x,
                self.DY,
                label_width_mark_x,
                self.DY + label_height,
            ),
            fill=mark_color,
        )

        labels = [
            # payload width
            {
                "xy": (self.X_MARGIN_PX + label_width / 2, preview_width_mark_y),
                "text": f"{px_to_mm(label_width - x_margin * 2)} mm",
                "anchor": "mm",
                "align": "center",
            },
            # label width
            {
                "xy": (self.X_MARGIN_PX + label_width / 2, label_width_mark_y),
                "text": f"{px_to_mm(label_width)} mm",
                "anchor": "mm",
                "align": "center",
            },
            # payload height
            {
                "xy": (preview_width_mark_x, self.DY + label_height / 2 - self.DY),
                "text": f"{px_to_mm(label_height - y_margin * 2)} mm",
                "anchor": "mm",
                "align": "center",
            },
            # label height
            {
                "xy": (label_width_mark_x, self.DY + label_height / 2 + self.DY),
                "text": f"{px_to_mm(label_height)} mm",
                "anchor": "mm",
                "align": "center",
            },
        ]
        for label in labels:
            bbox = draw.textbbox(**label)  # type: ignore[arg-type]
            draw.rectangle(bbox, fill=(0, 0, 0, 0))
            draw.text(**label, fill=text_color)  # type: ignore[arg-type]

    def render(self, context: RenderContext) -> Image.Image:
        label_bitmap, meta = self._get_label_bitmap(context)
        if context.preview_show_margins:
            label_width, label_height = label_bitmap.size
            preview_width = label_width + self.X_MARGIN_PX + self.DX
            preview_height = label_height + self.Y_MARGIN_PX + self.DY
            preview_bitmap = Image.new(
                "RGBA",
                (round(preview_width), round(preview_height)),
                color=(0, 0, 0, 0),
            )
            preview_bitmap.paste(label_bitmap, box=(self.X_MARGIN_PX, round(self.DY)))
            self._show_margins(label_bitmap, preview_bitmap, meta, context)
            return preview_bitmap
        else:
            return label_bitmap



from __future__ import annotations

from pathlib import Path

from PIL import Image

from labelle.lib.constants import DEFAULT_BARCODE_TYPE, BarcodeType, Direction
from labelle.lib.render_engines.barcode import BarcodeRenderEngine
from labelle.lib.render_engines.render_context import RenderContext
from labelle.lib.render_engines.render_engine import RenderEngine
from labelle.lib.render_engines.text import TextRenderEngine


class BarcodeWithTextRenderEngine(RenderEngine):
    TEXT_HEIGHT_SCALE_FACTOR = 0.4

    def __init__(
        self,
        content: str,
        font_file_name: Path | str,
        barcode_type: BarcodeType = DEFAULT_BARCODE_TYPE,
        frame_width_px: int = 0,
        font_size_ratio: float = 0.9,
        align: Direction = Direction.CENTER,
    ):
        super().__init__()
        self._barcode = BarcodeRenderEngine(content, barcode_type)
        self._text = TextRenderEngine(
            content, font_file_name, frame_width_px, font_size_ratio, align
        )
        self.align = align

    def render(self, render_context: RenderContext) -> Image.Image:
        bitmap = self._barcode.render(render_context)
        text_render_context = RenderContext(
            height_px=int(render_context.height_px * self.TEXT_HEIGHT_SCALE_FACTOR),
            foreground_color=render_context.foreground_color,
            background_color=render_context.background_color,
        )
        text_bitmap = self._text.render(text_render_context)

        # Define the x and y of the upper-left corner of the text
        # to be pasted onto the barcode
        text_offset_x = bitmap.height - text_bitmap.height - 1
        if self.align == Direction.LEFT:
            text_offset_y = 0
        elif self.align == Direction.CENTER:
            text_offset_y = bitmap.width // 2 - text_bitmap.width // 2
        elif self.align == Direction.RIGHT:
            text_offset_y = bitmap.width - text_bitmap.width

        bitmap.paste(text_bitmap, (text_offset_y, text_offset_x))
        return bitmap



from __future__ import annotations

from abc import ABC, abstractmethod

from PIL import Image

from labelle.lib.render_engines.render_context import RenderContext


class RenderEngineException(Exception):
    pass


class RenderEngine(ABC):
    @abstractmethod
    def render(self, context: RenderContext) -> Image.Image:
        raise NotImplementedError()

    def render_with_meta(
        self, context: RenderContext
    ) -> tuple[Image.Image, dict[str, float] | None]:
        return self.render(context), None



from __future__ import annotations

import logging

from PyQt6 import QtCore
from PyQt6.QtWidgets import (
    QComboBox,
    QLabel,
    QToolBar,
)

from labelle.lib.devices.online_device_manager import OnlineDeviceManager
from labelle.lib.devices.usb_device import UsbDevice

LOG = logging.getLogger(__name__)


class QDeviceSelector(QToolBar):
    _device_manager: OnlineDeviceManager
    _selected_device: UsbDevice | None

    selectedDeviceChangedSignal = QtCore.pyqtSignal(name="selectedDeviceChangedSignal")

    def __init__(self, parent=None):
        super().__init__(parent)
        self._devices = QComboBox(self)
        self._error_label = QLabel(self)

        self._selected_device = None
        self._action_devices = None
        self._action_error_label = None

        self._init_elements()
        self._init_connections()
        self._init_layout()

        self._last_scan_error_changed()
        self.selectedDeviceChangedSignal.emit()

    def _init_elements(self) -> None:
        self.device_manager = OnlineDeviceManager()

    def _init_connections(self) -> None:
        self.device_manager.devices_changed_signal.connect(self.repopulate)
        self.device_manager.last_scan_error_changed_signal.connect(
            self._last_scan_error_changed
        )
        self._devices.currentIndexChanged.connect(self._index_changed)

    def repopulate(self) -> None:
        old_hashes = {device.hash for device in self.device_manager.devices}
        self._devices.clear()
        for idx, device in enumerate(self.device_manager.devices):
            self._devices.insertItem(idx, device.product, device.hash)
            if (
                self._selected_device is not None
                and self._selected_device.hash == device.hash
            ):
                self._devices.setCurrentIndex(idx)
        valid = len(self.device_manager.devices) > 0
        if valid:
            if self._selected_device is None:
                self._index_changed(0)
        else:
            self._index_changed(-1)
        assert self._action_devices is not None
        assert self._action_error_label is not None
        self._action_devices.setVisible(valid)
        self._action_error_label.setVisible(not valid)
        new_hashes = {device.hash for device in self.device_manager.devices}
        if new_hashes != old_hashes:
            self.selectedDeviceChangedSignal.emit()

    def _index_changed(self, index) -> None:
        if index >= 0:
            self._selected_device = self.device_manager.devices[index]
        else:
            self._selected_device = None
        self.selectedDeviceChangedSignal.emit()

    def _last_scan_error_changed(self) -> None:
        last_scan_error = self.device_manager.last_scan_error or ""
        self._error_label.setText(str(last_scan_error))
        self.repopulate()

    def _init_layout(self) -> None:
        self._devices.setSizeAdjustPolicy(QComboBox.SizeAdjustPolicy.AdjustToContents)
        self._action_devices = self.addWidget(self._devices)
        self._action_error_label = self.addWidget(self._error_label)

    @property
    def selected_device(self) -> UsbDevice | None:
        device = None
        if self._devices.currentIndex() >= 0:
            device = self.device_manager.devices[self._devices.currentIndex()]
        return device



# === LICENSE STATEMENT ===
# Copyright (c) 2011 Sebastian J. Bronner <waschtl@sbronner.com>
#
# Copying and distribution of this file, with or without modification, are
# permitted in any medium without royalty provided the copyright notice and
# this notice are preserved.
# === END LICENSE STATEMENT ===

from typing import List, Tuple, Union

from PIL import Image, ImageDraw

from labelle.lib.barcode_writer import BinaryString


def _mm2px(mm: float, dpi: float = 25.4) -> float:
    return (mm * dpi) / 25.4


def _list_of_runs(line: BinaryString) -> List[int]:
    # Pack line to list give better gfx result, otherwise in can
    # result in aliasing gaps
    # '11010111' -> [2, -1, 1, -1, 3]
    c = 1
    mlist = []
    for i in range(0, len(line)):
        if i + 1 < len(line) and line[i] == line[i + 1]:
            c += 1
        else:
            if line[i] == "1":
                mlist.append(c)
            else:
                mlist.append(-c)
            c = 1
    return mlist


def _calculate_size(
    *,
    modules_per_line: int,
    quiet_zone: float,
    module_width: float,
    module_height: float,
    vertical_margin: float,
    dpi: float = 25.4,
) -> Tuple[int, int]:
    width = 2 * quiet_zone + modules_per_line * module_width
    height = vertical_margin * 2 + module_height
    return int(_mm2px(width, dpi)), int(_mm2px(height, dpi))


def convert_binary_string_to_barcode_image(
    line: BinaryString, quiet_zone: float, module_height: float
) -> Image.Image:
    """Render a barcode string into an image.

    line: A string of 0s and 1s representing the barcode.
    """
    module_width = 2
    vertical_margin = 8
    dpi = 25.4

    width, height = _calculate_size(
        modules_per_line=len(line),
        dpi=dpi,
        quiet_zone=quiet_zone,
        module_width=module_width,
        module_height=module_height,
        vertical_margin=vertical_margin,
    )
    image = Image.new("1", (width, height), 0)
    draw = ImageDraw.Draw(image)

    ypos = vertical_margin
    mlist = _list_of_runs(line)
    # Left quiet zone is x startposition
    xpos = quiet_zone
    for mod in mlist:
        if mod < 1:
            color = 0
        else:
            color = 1
        # remove painting for background colored tiles?
        _paint_module(
            xpos=xpos,
            ypos=ypos,
            width=module_width * abs(mod),
            color=color,
            dpi=dpi,
            module_height=module_height,
            draw=draw,
        )
        xpos += module_width * abs(mod)
    return image


def _paint_module(
    *,
    xpos: float,
    ypos: float,
    width: float,
    color: Union[int, str],
    dpi: float,
    module_height: float,
    draw: ImageDraw.ImageDraw,
) -> None:
    size = (
        (_mm2px(xpos, dpi), _mm2px(ypos, dpi)),
        (
            _mm2px(xpos + width, dpi),
            _mm2px(ypos + module_height, dpi),
        ),
    )
    draw.rectangle(size, outline=color, fill=color)



from __future__ import annotations

import logging

from usb.core import NoBackendError, USBError

from labelle.lib.constants import (
    SUPPORTED_PRODUCTS,
    UNCONFIRMED_MESSAGE,
)
from labelle.lib.devices.usb_device import UsbDevice

LOG = logging.getLogger(__name__)
POSSIBLE_USB_ERRORS = (NoBackendError, USBError)


class DeviceManagerError(RuntimeError):
    pass


class DeviceManagerNoDevices(DeviceManagerError):
    pass


class DeviceManager:
    _devices: dict[str, UsbDevice]

    def __init__(self) -> None:
        self._devices = {}

    def scan(self) -> bool:
        prev = self._devices
        try:
            cur = {dev.hash: dev for dev in UsbDevice.supported_devices() if dev.hash}
        except POSSIBLE_USB_ERRORS as e:
            self._devices.clear()
            raise DeviceManagerError(f"Failed scanning devices: {e}") from e
        if len(cur) == 0:
            self._devices.clear()
            raise DeviceManagerNoDevices("No supported devices found")

        prev_set = set(prev)
        cur_set = set(cur)

        for dev in prev_set - cur_set:
            self._devices.pop(dev)
        for dev in cur_set - prev_set:
            self._devices[dev] = cur[dev]

        changed = prev_set != cur_set
        return changed

    @property
    def devices(self) -> list[UsbDevice]:
        try:
            return sorted(self._devices.values(), key=lambda dev: dev.hash)
        except POSSIBLE_USB_ERRORS:
            return []

    def matching_devices(self, patterns: list[str] | None) -> list[UsbDevice]:
        try:
            matching = filter(
                lambda dev: dev.is_match(patterns), self._devices.values()
            )
            return sorted(matching, key=lambda dev: dev.hash)
        except POSSIBLE_USB_ERRORS:
            return []

    def find_and_select_device(self, patterns: list[str] | None = None) -> UsbDevice:
        devices = [
            device for device in self.matching_devices(patterns) if device.is_supported
        ]
        if len(devices) == 0:
            raise DeviceManagerError("No matching devices found")
        if len(devices) > 1:
            LOG.debug("Found multiple matching Dymo devices. Using first device")
        else:
            LOG.debug("Found single device")
        for dev in devices:
            LOG.debug(dev.device_info)
        dev = devices[0]
        if dev.is_supported:
            msg = f"Recognized device as {SUPPORTED_PRODUCTS[dev.id_product]}"
        else:
            msg = f"Unrecognized device: {hex(dev.id_product)}. {UNCONFIRMED_MESSAGE}"
        LOG.debug(msg)
        return dev



from __future__ import annotations

import math
from pathlib import Path

from PIL import Image, ImageOps, UnidentifiedImageError

from labelle.lib.render_engines import NoContentError
from labelle.lib.render_engines.render_context import RenderContext
from labelle.lib.render_engines.render_engine import (
    RenderEngine,
    RenderEngineException,
)


class PicturePathDoesNotExist(RenderEngineException):
    pass


class UnidentifiedImageFileError(RenderEngineException):
    def __init__(self, exception) -> None:
        super().__init__(exception)


class PictureRenderEngine(RenderEngine):
    def __init__(self, picture_path: Path | str) -> None:
        super().__init__()
        if picture_path == "":
            raise NoContentError()
        self.picture_path = Path(picture_path)
        if not self.picture_path.is_file():
            raise PicturePathDoesNotExist(
                f"Picture path does not exist: {picture_path}"
            )

    def render(self, context: RenderContext) -> Image.Image:
        height_px = context.height_px
        try:
            with Image.open(self.picture_path) as img:
                if img.height > height_px:
                    ratio = height_px / img.height
                    img = img.resize((math.ceil(img.width * ratio), height_px))

                img = img.convert("L", palette=Image.AFFINE)
                return ImageOps.invert(img).convert("1")
        except UnidentifiedImageError as e:
            raise UnidentifiedImageFileError(e) from e



import logging
from typing import Optional

from PyQt6 import QtCore
from PyQt6.QtCore import QSize, Qt
from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import QLabel, QPushButton, QVBoxLayout, QWidget

LOG = logging.getLogger(__name__)


class QActions(QWidget):
    _error_label: QLabel
    _is_enabled: bool
    _last_error: Optional[str]
    _print_button: QPushButton

    print_label_signal = QtCore.pyqtSignal(name="printLabel")

    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self._error_label = QLabel()
        self._is_enabled = False
        self._last_error = None
        self._print_button = QPushButton()

        self._init_elements()
        self._init_connections()
        self._init_layout()

    def _init_elements(self) -> None:
        printer_icon = QIcon.fromTheme("printer")
        self._print_button.setIcon(printer_icon)
        self._print_button.setFixedSize(64, 64)
        self._print_button.setIconSize(QSize(48, 48))

    def _init_connections(self) -> None:
        self._print_button.clicked.connect(self._on_print_label)

    def _init_layout(self) -> None:
        layout = QVBoxLayout(self)
        layout.addWidget(
            self._print_button, alignment=QtCore.Qt.AlignmentFlag.AlignRight
        )
        layout.addWidget(
            self._error_label, alignment=QtCore.Qt.AlignmentFlag.AlignCenter
        )

    def _on_print_label(self) -> None:
        self.print_label_signal.emit()

    def clear_error(self) -> None:
        self._error_label.setText("")
        self._last_error = None
        self._print_button.setEnabled(True)
        self._print_button.setCursor(Qt.CursorShape.ArrowCursor)

    def set_error(self, error: str) -> None:
        if self._last_error == error:
            return
        self._last_error = error
        self._error_label.setText(error)
        LOG.error(error)
        self._print_button.setDisabled(True)
        self._print_button.setCursor(Qt.CursorShape.ForbiddenCursor)



# === LICENSE STATEMENT ===
# Copyright (c) 2011 Sebastian J. Bronner <waschtl@sbronner.com>
#
# Copying and distribution of this file, with or without modification, are
# permitted in any medium without royalty provided the copyright notice and
# this notice are preserved.
# === END LICENSE STATEMENT ===

from typing import List, NamedTuple, NewType

from barcode.writer import BaseWriter

BinaryString = NewType("BinaryString", str)
"""A string that's been validated to contain only '0's and '1's."""


def _noop() -> None:
    pass


def _validate_string_as_binary(s: str) -> BinaryString:
    if not all(c in ("0", "1") for c in s):
        raise ValueError("Barcode can only contain 0 and 1")
    return BinaryString(s)


class BarcodeResult(NamedTuple):
    line: BinaryString
    quiet_zone: float


class SimpleBarcodeWriter(BaseWriter):
    def __init__(self) -> None:
        # Pass dummy values because we re-implement the `render` method anyway
        super().__init__(
            initialize=None,
            paint_module=_noop,
            paint_text=None,
            finish=_noop,
        )

    def render(self, code: List[str]) -> BarcodeResult:
        """Extract the barcode string from the code and render it into an image."""
        if len(code) != 1:
            raise ValueError("Barcode expected to have only one line")
        line = _validate_string_as_binary(code[0])
        return BarcodeResult(line=line, quiet_zone=self.quiet_zone)



from __future__ import annotations

import logging
import platform
from textwrap import dedent
from typing import Any, NoReturn

import usb

from labelle.lib.constants import (
    DEV_VENDOR,
    HID_INTERFACE_CLASS,
    PRINTER_INTERFACE_CLASS,
    SUPPORTED_PRODUCTS,
)

LOG = logging.getLogger(__name__)
GITHUB_ISSUE_MAC = "<https://github.com/labelle-org/labelle/issues/5>"
GITHUB_ISSUE_UDEV = "<https://github.com/labelle-org/labelle/issues/6>"


class UsbDeviceError(RuntimeError):
    pass


class UsbDevice:
    _dev: usb.core.Device
    _intf: usb.core.Interface | None
    _devin: usb.core.Endpoint | None
    _devout: usb.core.Endpoint | None

    def __init__(self, dev: usb.core.Device) -> None:
        self._dev = dev
        self._intf = None
        self._devin = None
        self._devout = None

    @property
    def hash(self) -> str:
        return self.usb_id

    def _get_dev_attribute(self, attr: str) -> Any:
        try:
            return getattr(self._dev, attr)
        except (ValueError, usb.core.USBError):
            return None

    @property
    def manufacturer(self) -> str | None:
        return self._get_dev_attribute("manufacturer")

    @property
    def product(self) -> str | None:
        return self._get_dev_attribute("product")

    @property
    def serial_number(self) -> str | None:
        return self._get_dev_attribute("serial_number")

    @property
    def id_vendor(self) -> int:
        id_ = self._get_dev_attribute("idVendor")
        if id_ is None:
            raise UsbDeviceError("Could not get idVendor")
        return id_

    @property
    def id_product(self) -> int:
        id_ = self._get_dev_attribute("idProduct")
        if id_ is None:
            raise UsbDeviceError("Could not get idProduct")
        return id_

    @property
    def vendor_product_id(self) -> str:
        return f"{self.id_vendor:04x}:{self.id_product:04x}"

    @property
    def usb_id(self) -> str:
        bus = self._get_dev_attribute("bus")
        address = self._get_dev_attribute("address")
        return f"Bus {bus:03} Device {address:03}: ID {self.vendor_product_id}"

    @staticmethod
    def _is_supported_vendor(dev: usb.core.Device) -> bool:
        return dev.idVendor == DEV_VENDOR

    @property
    def is_supported(self) -> bool:
        return (
            self._is_supported_vendor(self._dev)
            and self.id_product in SUPPORTED_PRODUCTS
        )

    @staticmethod
    def supported_devices() -> set[UsbDevice]:
        return {
            UsbDevice(dev)
            for dev in usb.core.find(
                find_all=True, custom_match=UsbDevice._is_supported_vendor
            )
        }

    @property
    def device_info(self) -> str:
        try:
            _ = self._dev.manufacturer
        except ValueError:
            self._instruct_on_access_denied()
        res = ""
        res += f"{self._dev!r}\n"
        res += f"  manufacturer: {self._dev.manufacturer}\n"
        res += f"  product: {self._dev.product}\n"
        res += f"  serial: {self._dev.serial_number}\n"
        configs = self._dev.configurations()
        if configs:
            res += "  configurations:\n"
            for cfg in configs:
                res += f"  - {cfg!r}\n"
                intfs = cfg.interfaces()
                if intfs:
                    res += "    interfaces:\n"
                    for intf in intfs:
                        res += f"    - {intf!r}\n"
        return res

    def _instruct_on_access_denied(self) -> NoReturn:
        system = platform.system()
        if system == "Linux":
            self._instruct_on_access_denied_linux()
        elif system == "Windows":
            raise UsbDeviceError(
                "Couldn't access the device. Please make sure that the "
                "device driver is set to WinUSB. This can be accomplished "
                "with Zadig <https://zadig.akeo.ie/>."
            )
        elif system == "Darwin":
            raise UsbDeviceError(
                f"Could not access {self._dev}. Thanks for bravely trying this on a "
                f"Mac. You are in uncharted territory. It would be appreciated if you "
                f"share the results of your experimentation at {GITHUB_ISSUE_MAC}."
            )
        else:
            raise UsbDeviceError(f"Unknown platform {system}")

    def _instruct_on_access_denied_linux(self) -> NoReturn:
        # try:
        #     os_release = platform.freedesktop_os_release()
        # except OSError:
        #     os_release = {}
        # dists_with_empties = [os_release.get("ID", "")] + os_release.get(
        #     "ID_LIKE", ""
        # ).split(" ")
        # dists = [dist for dist in dists_with_empties if dist]
        # if "arch" in dists:
        #     restart_udev_command = "sudo udevadm control --reload"
        # elif "ubuntu" in dists or "debian" in dists:
        #     restart_udev_command = "sudo systemctl restart udev.service"
        # # detect whether we are in arch linux or ubuntu linux
        # if Path("/etc/arch-release").exists():
        #     restart_udev_command = "sudo udevadm control --reload"
        # elif Path("/etc/lsb-release").exists():
        #     restart_udev_command = "sudo systemctl restart udev.service"
        # else:
        #     restart_udev_command = None

        udev_rule = ",'\\\n                     '".join(
            [
                'ACTION=="add"',
                'SUBSYSTEMS=="usb"',
                f'ATTRS{{idVendor}}=="{self.id_vendor:04x}"',
                f'ATTRS{{idProduct}}=="{self.id_product:04x}"',
                'MODE="0666"',
            ]
        )

        msg = f"""

            You do not have sufficient access to the device. You probably want to add the a udev
            rule in /etc/udev/rules.d with the following command:

                echo '{udev_rule}' \\
                  | sudo tee /etc/udev/rules.d/91-labelle-{self.id_vendor:04x}-{self.id_product:04x}.rules

            Next, refresh udev with:

                sudo udevadm control --reload-rules
                sudo udevadm trigger --attr-match=idVendor="0922"

            Finally, turn your device off and back on again to activate the new permissions.

            If this still does not resolve the problem, you might need to reboot.
            In case rebooting is necessary, please report this at {GITHUB_ISSUE_UDEV}.
            We are still trying to figure out a simple procedure which works for everyone.
            In case you still cannot connect, or if you have any information or ideas, please
            post them at that link.
        """  # noqa: E501
        msg = dedent(msg)
        raise UsbDeviceError(msg)

    def _set_configuration(self) -> None:
        try:
            self._dev.get_active_configuration()
            LOG.debug("Active device configuration already found.")
        except usb.core.USBError:
            try:
                self._dev.set_configuration()
                LOG.debug("Device configuration set.")
            except usb.core.USBError as e:
                if e.errno == 13:
                    raise UsbDeviceError("Access denied") from e
                if e.errno == 16:
                    LOG.debug("Device is busy, but this is okay.")
                else:
                    raise

    def setup(self) -> None:
        try:
            self._setup()
        except usb.core.USBError as e:
            raise UsbDeviceError(f"Failed setup USB device: {e}") from e

    def _setup(self):
        self._set_configuration()
        intf = usb.util.find_descriptor(
            self._dev.get_active_configuration(),
            bInterfaceClass=PRINTER_INTERFACE_CLASS,
        )
        if intf is not None:
            LOG.debug(f"Opened printer interface: {intf!r}")
        else:
            intf = usb.util.find_descriptor(
                self._dev.get_active_configuration(),
                bInterfaceClass=HID_INTERFACE_CLASS,
            )
            if intf is not None:
                LOG.debug(f"Opened HID interface: {intf!r}")
            else:
                raise UsbDeviceError("Could not open a valid interface")
        assert isinstance(intf, usb.core.Interface)

        try:
            if self._dev.is_kernel_driver_active(intf.bInterfaceNumber):
                LOG.debug(
                    f"Detaching kernel driver from interface {intf.bInterfaceNumber}"
                )
                self._dev.detach_kernel_driver(intf.bInterfaceNumber)
        except NotImplementedError:
            LOG.debug(f"Kernel driver detaching not necessary on {platform.system()}.")
        devout = usb.util.find_descriptor(
            intf,
            custom_match=(
                lambda e: usb.util.endpoint_direction(e.bEndpointAddress)
                == usb.util.ENDPOINT_OUT
            ),
        )
        devin = usb.util.find_descriptor(
            intf,
            custom_match=(
                lambda e: usb.util.endpoint_direction(e.bEndpointAddress)
                == usb.util.ENDPOINT_IN
            ),
        )

        if not devout or not devin:
            self._intf = None
            self._devin = None
            self._devout = None
            raise UsbDeviceError("The device endpoints not be found")
        self._intf = intf
        self._devin = devin
        self._devout = devout

    def dispose(self) -> None:
        usb.util.dispose_resources(self._dev)

    def is_match(self, patterns: list[str] | None) -> bool:
        if patterns is None:
            return True
        match = True
        for pattern in patterns:
            pattern = pattern.lower()
            match &= (
                pattern in (self.manufacturer or "").lower()
                or pattern in (self.product or "").lower()
                or pattern in (self.serial_number or "").lower()
            )
        return match

    @property
    def devin(self):
        return self._devin

    @property
    def devout(self):
        return self._devout



# === LICENSE STATEMENT ===
# Copyright (c) 2011 Sebastian J. Bronner <waschtl@sbronner.com>
#
# Copying and distribution of this file, with or without modification, are
# permitted in any medium without royalty provided the copyright notice and
# this notice are preserved.
# === END LICENSE STATEMENT ===

# On systems with access to sysfs under /sys, this script will use the three
# variables DEV_CLASS, DEV_VENDOR, and DEV_PRODUCT to find the device file
# under /dev automatically. This behavior can be overridden by setting the
# variable DEV_NODE to the device file path. This is intended for cases, where
# either sysfs is unavailable or unusable by this script for some reason.
# Please beware that DEV_NODE must be set to None when not used, else you will
# be bitten by the NameError exception.
from enum import Enum
from pathlib import Path

import labelle.resources.fonts
import labelle.resources.icons

try:
    from pyqrcode import QRCode

    USE_QR = True
    e_qrcode = None
except ImportError as error:  # pragma: no cover
    e_qrcode = error
    USE_QR = False
    QRCode = None


UNCONFIRMED_MESSAGE = (
    "WARNING: This device is not confirmed to work with this software. Please "
    "report your experiences in https://github.com/labelle-org/labelle/issues/4"
)
SUPPORTED_PRODUCTS = {
    0x0011: "DYMO LabelMANAGER PC",
    0x0015: "LabelPoint 350",
    0x1001: "LabelManager PnP (no mode switch)",
    0x1002: "LabelManager PnP (mode switch)",
    0x1003: f"LabelManager 420P (no mode switch) {UNCONFIRMED_MESSAGE}",
    0x1004: f"LabelManager 420P (mode switch) {UNCONFIRMED_MESSAGE}",
    0x1005: "LabelManager 280 (no mode switch)",
    0x1006: "LabelManager 280 (no mode switch)",
    0x1007: f"LabelManager Wireless PnP (no mode switch) {UNCONFIRMED_MESSAGE}",
    0x1008: f"LabelManager Wireless PnP (mode switch) {UNCONFIRMED_MESSAGE}",
    0x1009: f"MobileLabeler {UNCONFIRMED_MESSAGE}",
}
DEV_VENDOR = 0x0922

PRINTER_INTERFACE_CLASS = 0x07
HID_INTERFACE_CLASS = 0x03

# Escape character preceeding all commands
ESC = 0x1B

# Synchronization character preceding uncompressed print data
SYN = 0x16

FONT_SIZERATIO = 7 / 8

DEFAULT_MARGIN_PX = 56
VERTICAL_PREVIEW_MARGIN_PX = 13

DPI = 180
MM_PER_INCH = 25.4
PIXELS_PER_MM = DPI / MM_PER_INCH

ICON_DIR = Path(labelle.resources.icons.__file__).parent


class BarcodeType(str, Enum):
    CODE128 = "code128"
    CODE39 = "code39"
    CODABAR = "codabar"
    EAN = "ean"
    EAN13 = "ean13"
    EAN13_GUARD = "ean13-guard"
    EAN14 = "ean14"
    EAN8 = "ean8"
    EAN8_GUARD = "ean8-guard"
    GS1 = "gs1"
    GS1_128 = "gs1_128"
    GTIN = "gtin"
    ISBN = "isbn"
    ISBN10 = "isbn10"
    ISBN13 = "isbn13"
    ISSN = "issn"
    ITF = "itf"
    JAN = "jan"
    NW_7 = "nw-7"
    PZN = "pzn"
    UPC = "upc"
    UPCA = "upca"


DEFAULT_BARCODE_TYPE = BarcodeType.CODE128


class Direction(str, Enum):
    LEFT = "left"
    CENTER = "center"
    RIGHT = "right"


class Output(str, Enum):
    BROWSER = "browser"
    CONSOLE = "console"
    CONSOLE_INVERTED = "console-inverted"
    IMAGEMAGICK = "imagemagick"
    PNG = "png"
    PRINTER = "printer"



from __future__ import annotations

from pathlib import Path

from PIL import Image, ImageFont

from labelle.lib.constants import Direction
from labelle.lib.render_engines.render_context import RenderContext
from labelle.lib.render_engines.render_engine import RenderEngine
from labelle.lib.utils import draw_image


class TextRenderEngine(RenderEngine):
    def __init__(
        self,
        text_lines: str | list[str],
        font_file_name: Path | str,
        frame_width_px: int = 0,
        font_size_ratio: float = 0.9,
        align: Direction = Direction.CENTER,
    ):
        if isinstance(text_lines, str):
            text_lines = [text_lines]

        if len(text_lines) == 0:
            text_lines = [" "]

        self.text_lines = text_lines
        self.font_file_name = font_file_name
        self.frame_width_px = frame_width_px
        self.font_size_ratio = font_size_ratio
        self.align = align

        super().__init__()

    def render(self, context: RenderContext) -> Image.Image:
        height_px = context.height_px
        line_height = float(height_px) / len(self.text_lines)
        font_size_px = round(line_height * self.font_size_ratio)

        font_offset_px = int((line_height - font_size_px) / 2)
        if self.frame_width_px:
            frame_width_px = self.frame_width_px or min(
                self.frame_width_px, font_offset_px, 3
            )
        else:
            frame_width_px = self.frame_width_px

        font = ImageFont.truetype(str(self.font_file_name), font_size_px)
        boxes = (font.getbbox(line) for line in self.text_lines)
        line_widths = (right - left for left, _top, right, _bottom in boxes)
        label_width_px = max(line_widths) + (font_offset_px * 2)
        bitmap = Image.new("1", (label_width_px, height_px))
        with draw_image(bitmap) as draw:
            # draw frame into empty image
            if frame_width_px:
                draw.rectangle(((0, 4), (label_width_px - 1, height_px - 4)), fill=1)
                draw.rectangle(
                    (
                        (frame_width_px, 4 + frame_width_px),
                        (
                            label_width_px - (frame_width_px + 1),
                            height_px - (frame_width_px + 4),
                        ),
                    ),
                    fill=0,
                )

            # write the text into the empty image
            multiline_text = "\n".join(self.text_lines)
            draw.multiline_text(
                (label_width_px / 2, height_px / 2),
                multiline_text,
                align=self.align.value,
                anchor="mm",
                font=font,
                fill=1,
            )
        return bitmap



from __future__ import annotations

from typing import Sequence

from PIL import Image

from labelle.lib.render_engines.empty import EmptyRenderEngine
from labelle.lib.render_engines.render_context import RenderContext
from labelle.lib.render_engines.render_engine import RenderEngine


class HorizontallyCombinedRenderEngine(RenderEngine):
    padding: int

    def __init__(
        self,
        render_engines: Sequence[RenderEngine],
        padding: int = 4,
    ):
        super().__init__()
        self.render_engines = render_engines
        self.padding = padding

    def render(self, context: RenderContext) -> Image.Image:
        render_engines = self.render_engines or [EmptyRenderEngine()]
        bitmaps = [engine.render(context) for engine in render_engines]

        if len(bitmaps) == 1:
            merged_bitmap = bitmaps[0]
        else:
            label_height = max(b.height for b in bitmaps)
            total_width = _compute_total_width(bitmaps, self.padding)
            merged_bitmap = Image.new("1", (total_width, label_height))
            x_offset = 0
            for bitmap in bitmaps:
                y_offset = (label_height - bitmap.height) // 2
                merged_bitmap.paste(bitmap, box=(x_offset, y_offset))
                x_offset += bitmap.width + self.padding

        return merged_bitmap


def _compute_total_width(bitmaps: Sequence[Image.Image], padding: int) -> int:
    number_of_bitmaps = len(bitmaps)
    width_of_bitmaps = sum(b.width for b in bitmaps)

    # We alternate bitmaps and padding, and have bitmaps on each end.
    number_of_paddings = number_of_bitmaps - 1
    width_of_paddings = number_of_paddings * padding

    total_width = width_of_bitmaps + width_of_paddings
    return total_width



import logging
import sys
from typing import Optional

from PIL import Image
from PyQt6 import QtCore
from PyQt6.QtCore import QCommandLineOption, QCommandLineParser
from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import QApplication, QHBoxLayout, QVBoxLayout, QWidget

from labelle.gui.common import crash_msg_box
from labelle.gui.q_actions import QActions
from labelle.gui.q_device_selector import QDeviceSelector
from labelle.gui.q_labels_list import QLabelList
from labelle.gui.q_render import QRender
from labelle.gui.q_settings_toolbar import QSettingsToolbar, Settings
from labelle.lib.constants import ICON_DIR
from labelle.lib.devices.device_manager import DeviceManager
from labelle.lib.devices.dymo_labeler import DymoLabeler, DymoLabelerPrintError
from labelle.lib.env_config import is_verbose_env_vars
from labelle.lib.logger import configure_logging, set_not_verbose
from labelle.lib.render_engines import RenderContext
from labelle.lib.utils import system_run

LOG = logging.getLogger(__name__)


class LabelleWindow(QWidget):
    _label_bitmap_to_print: Optional[Image.Image]
    _device_manager: DeviceManager
    _dymo_labeler: DymoLabeler
    _render_context: RenderContext
    _render_widget: QWidget

    def __init__(self) -> None:
        super().__init__()
        self._label_bitmap_to_print = None
        self._detected_device = None

        self._window_layout = QVBoxLayout()

        self._device_selector = QDeviceSelector(self)
        self._label_list = QLabelList()
        self._render = QRender(self)
        self._actions = QActions(self)
        self._settings_toolbar = QSettingsToolbar(self)
        self._render_widget = QWidget(self)

        self._init_elements()
        self._init_connections()
        self._init_layout()

        self._device_selector.repopulate()
        self._settings_toolbar.on_settings_changed()
        self._label_list.populate()
        self._label_list.render_label()

    def _init_elements(self) -> None:
        self.setWindowTitle("Labelle GUI")
        self.setWindowIcon(QIcon(str(ICON_DIR / "logo_small.png")))
        self.setGeometry(200, 200, 1100, 400)

        self._device_manager = DeviceManager()
        self._dymo_labeler = DymoLabeler()
        self._settings_toolbar.update_labeler_context(
            supported_tape_sizes=self._dymo_labeler.SUPPORTED_TAPE_SIZES_MM,
            installed_tape_size=self._dymo_labeler.tape_size_mm,
            minimum_horizontal_margin_mm=self._dymo_labeler.minimum_horizontal_margin_mm,
        )

    def _init_connections(self) -> None:
        self._label_list.renderPrintPreviewSignal.connect(self._update_preview_render)
        self._label_list.renderPrintPayloadSignal.connect(self._update_print_render)
        self._actions.print_label_signal.connect(self._on_print_label)
        self._settings_toolbar.settings_changed_signal.connect(
            self._on_settings_changed
        )
        self._device_selector.selectedDeviceChangedSignal.connect(
            self._on_device_selected
        )

    def _init_layout(self) -> None:
        self._actions.setParent(self._render_widget)
        self._render.setParent(self._render_widget)

        render_layout = QHBoxLayout(self._render_widget)
        render_layout.addWidget(
            self._render, alignment=QtCore.Qt.AlignmentFlag.AlignRight
        )
        render_layout.addWidget(
            self._actions, alignment=QtCore.Qt.AlignmentFlag.AlignRight
        )

        self._window_layout.addWidget(self._device_selector)
        self._window_layout.addWidget(self._settings_toolbar)
        self._window_layout.addWidget(self._label_list)
        self._window_layout.addWidget(self._render_widget)
        self.setLayout(self._window_layout)

    def _on_settings_changed(self, settings: Settings) -> None:
        assert self._dymo_labeler is not None
        self._dymo_labeler.tape_size_mm = settings.tape_size_mm

        # Update render context
        self._render_context = RenderContext(
            foreground_color=settings.foreground_color,
            background_color=settings.background_color,
            height_px=self._dymo_labeler.height_px,
            preview_show_margins=settings.preview_show_margins,
        )
        self._label_list.update_params(
            dymo_labeler=self._dymo_labeler,
            h_margin_mm=settings.horizontal_margin_mm,
            min_label_width_mm=settings.min_label_width_mm,
            render_context=self._render_context,
            justify=settings.justify,
        )

        is_ready = self._dymo_labeler.is_ready
        self._settings_toolbar.setEnabled(is_ready)
        self._label_list.setEnabled(is_ready)
        self._render_widget.setEnabled(is_ready)

    def _update_preview_render(self, preview_bitmap: Image.Image) -> None:
        self._render.update_preview_render(preview_bitmap)

    def _update_print_render(self, label_bitmap_to_print) -> None:
        self._label_bitmap_to_print = label_bitmap_to_print

    def _on_print_label(self) -> None:
        try:
            if self._label_bitmap_to_print is None:
                raise RuntimeError("No label to print! Call update_label_render first.")
            assert self._dymo_labeler is not None
            self._dymo_labeler.print(self._label_bitmap_to_print)
        except DymoLabelerPrintError as err:
            crash_msg_box(self, "Printing Failed!", err)

    def _on_device_selected(self) -> None:
        self._dymo_labeler.device = self._device_selector.selected_device
        self._settings_toolbar.on_settings_changed()


def parse(app) -> None:
    """Parse the arguments and options of the given app object."""
    parser = QCommandLineParser()
    parser.addHelpOption()

    verbose_option = QCommandLineOption(["v", "verbose"], "Verbose output.")
    parser.addOption(verbose_option)
    parser.process(app)

    is_verbose = parser.isSet(verbose_option)
    if (not is_verbose) and (not is_verbose_env_vars()):
        # Neither the --verbose flag nor the environment variable is set.
        set_not_verbose()


def main() -> None:
    configure_logging()
    with system_run():
        app = QApplication(sys.argv)
        parse(app)
        window = LabelleWindow()
        window.show()
        sys.exit(app.exec())


if __name__ == "__main__":
    main()



from PIL import Image

from labelle.lib.render_engines.render_context import RenderContext
from labelle.lib.render_engines.render_engine import RenderEngine


class EmptyRenderEngine(RenderEngine):
    def __init__(self, width_px: int = 1):
        super().__init__()
        self.width_px = width_px

    def render(self, context: RenderContext) -> Image.Image:
        return Image.new("1", (self.width_px, context.height_px))



import logging
from typing import Optional

from PIL import Image, ImageQt
from PyQt6.QtGui import QPixmap
from PyQt6.QtWidgets import QGraphicsDropShadowEffect, QLabel, QWidget

LOG = logging.getLogger(__name__)


class QRender(QLabel):
    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)
        self._init_elements()

    def _init_elements(self) -> None:
        shadow = QGraphicsDropShadowEffect()
        shadow.setBlurRadius(15)
        self.setGraphicsEffect(shadow)

    def update_preview_render(self, preview_bitmap: Image.Image) -> None:
        qim = ImageQt.ImageQt(preview_bitmap)
        q_image = QPixmap.fromImage(qim)
        self.setPixmap(q_image)
        self.adjustSize()



import logging
import warnings
from configparser import ConfigParser
from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, Optional

from platformdirs import user_config_dir

logger = logging.getLogger(__name__)


def get_config_file() -> Path:
    old_config_file = Path(user_config_dir()) / "dymoprint.ini"
    config_file = Path(user_config_dir()) / "labelle.ini"
    if old_config_file.exists() and not config_file.exists():
        warnings.warn(
            f"Old config file found at {old_config_file}. "
            f"Please rename it to {config_file}",
            stacklevel=2,
        )
        config_file = old_config_file
    return config_file


@lru_cache
def get_config() -> ConfigParser:
    config_parser = ConfigParser()
    file_to_read = get_config_file()
    if config_parser.read(file_to_read):
        logger.debug(f"Read config file: {file_to_read}")
    else:
        logger.debug(f"Config file not found: {file_to_read}")
    return config_parser


def get_config_section(section_name: str) -> Optional[Dict[str, Any]]:
    config = get_config()
    if section_name not in config:
        return None
    return dict(config[section_name])



from __future__ import annotations

from pathlib import Path

from PyQt6 import QtCore
from PyQt6.QtGui import (
    QFont,
    QFontDatabase,
    QIcon,
    QStandardItem,
    QStandardItemModel,
)
from PyQt6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QFileDialog,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPlainTextEdit,
    QPushButton,
    QSpinBox,
    QWidget,
)

from labelle.gui.common import crash_msg_box
from labelle.lib.constants import ICON_DIR, BarcodeType, Direction
from labelle.lib.env_config import is_dev_mode_no_margins
from labelle.lib.font_config import get_available_fonts
from labelle.lib.render_engines import (
    BarcodeRenderEngine,
    BarcodeWithTextRenderEngine,
    EmptyRenderEngine,
    NoContentError,
    PicturePathDoesNotExist,
    PictureRenderEngine,
    QrRenderEngine,
    RenderContext,
    RenderEngine,
    TextRenderEngine,
)
from labelle.lib.render_engines.render_engine import RenderEngineException


class FontStyle(QComboBox):
    def __init__(self) -> None:
        super().__init__()

        # Populate font_style
        fonts_model = QStandardItemModel()

        # Create items for all available fonts
        for font_path in get_available_fonts():
            # Create item for font
            item = self.make_combobox_item_for_font(font_path)
            fonts_model.appendRow(item)

        self.setModel(fonts_model)

        # Select default font
        self.setCurrentText("Carlito-Regular")

    def make_combobox_item_for_font(self, font_path: Path) -> QStandardItem:
        # Retrieve font data
        font_name = font_path.stem
        font_absolute_path = font_path.absolute()

        # Make combobox item
        item = QStandardItem(font_name)
        item.setData(font_absolute_path, QtCore.Qt.ItemDataRole.UserRole)
        item_font = QFont()

        # Add application font to allow Qt rendering with it
        font_id = QFontDatabase.addApplicationFont(str(font_absolute_path))
        if font_id >= 0:
            loaded_font_families = QFontDatabase.applicationFontFamilies(font_id)
            if loaded_font_families:
                item_font.setFamilies(loaded_font_families)

        # Set bold if font name indictates it
        if "bold" in font_name.lower():
            item_font.setBold(True)

        # Set italic if font name indictates it
        if "italic" in font_name.lower():
            item_font.setItalic(True)

        # font to item
        item.setFont(item_font)

        return item


class BaseLabelWidget(QWidget):
    """A base class for creating Dymo label widgets.

    Signals:
    --------
    itemRenderSignal : PyQtSignal
        Signal emitted when the content of the label is changed.

    Methods
    -------
    content_changed()
        Emits the itemRenderSignal when the content of the label is changed.
    render_label()
        Abstract method to be implemented by subclasses for rendering the label.

    """

    render_context: RenderContext

    itemRenderSignal = QtCore.pyqtSignal(name="itemRenderSignal")

    def content_changed(self) -> None:
        """Emit the itemRenderSignal when the content of the label is changed."""
        self.itemRenderSignal.emit()

    @property
    def render_engine_impl(self) -> None:
        """Abstract method for getting the render engine of the label."""
        pass

    @property
    def render_engine(self) -> RenderEngine | None:
        try:
            return self.render_engine_impl
        except RenderEngineException as err:
            crash_msg_box(self, "Render Engine Failed!", err)
            return EmptyRenderEngine()


class TextDymoLabelWidget(BaseLabelWidget):
    """A widget for rendering text on a Dymo label.

    Args:
    ----
        render_context (RenderContext): The rendering context to use.
        parent (QWidget): The parent widget of this widget.

    Attributes:
    ----------
        render_context (RenderContext): The rendering context used by this widget.
        label (QPlainTextEdit): The text label to be rendered on the Dymo label.
        font_style (FontStyle): The font style selection dropdown.
        font_size (QSpinBox): The font size selection spinner.
        frame_width_px (QSpinBox): The frame width selection spinner.
    Signals:
        itemRenderSignal: A signal emitted when the content of the label changes.

    """

    align: QComboBox
    label: QPlainTextEdit
    font_style: FontStyle
    font_size: QSpinBox
    frame_width_px: QSpinBox

    def __init__(self, render_context: RenderContext, parent: QWidget | None = None):
        super().__init__(parent)
        self.render_context = render_context

        default_label_text = "." if is_dev_mode_no_margins() else "text"
        self.label = QPlainTextEdit(default_label_text)
        self.label.setFixedHeight(15 * (len(self.label.toPlainText().splitlines()) + 2))
        self.setFixedHeight(self.label.height() + 10)
        self.font_style = FontStyle()
        self.font_size = QSpinBox()
        self.font_size.setMaximum(150)
        self.font_size.setMinimum(0)
        self.font_size.setSingleStep(1)
        self.font_size.setValue(90)
        self.frame_width_px = QSpinBox()
        self.align = QComboBox()

        self.align.addItems(["left", "center", "right"])

        layout = QHBoxLayout()
        item_icon = QLabel()
        item_icon.setPixmap(QIcon(str(ICON_DIR / "txt_icon.png")).pixmap(32, 32))
        item_icon.setAlignment(QtCore.Qt.AlignmentFlag.AlignHCenter)
        layout.addWidget(item_icon)
        layout.addWidget(self.label)
        layout.addWidget(QLabel("Font:"))
        layout.addWidget(self.font_style)
        layout.addWidget(QLabel("Size [%]:"))
        layout.addWidget(self.font_size)
        layout.addWidget(QLabel("Frame Width:"))
        layout.addWidget(self.frame_width_px)
        layout.addWidget(QLabel("Alignment:"))
        layout.addWidget(self.align)
        self.label.textChanged.connect(self.content_changed)
        self.frame_width_px.valueChanged.connect(self.content_changed)
        self.font_size.valueChanged.connect(self.content_changed)
        self.font_style.currentTextChanged.connect(self.content_changed)
        self.align.currentTextChanged.connect(self.content_changed)
        self.setLayout(layout)

    def content_changed(self) -> None:
        """Manage changes to the label contents.

        In particular, update the height of the label and emit the itemRenderSignal
        when the content of the label changes.
        """
        self.label.setFixedHeight(15 * (len(self.label.toPlainText().splitlines()) + 2))
        self.setFixedHeight(self.label.height() + 10)
        self.itemRenderSignal.emit()

    @property
    def render_engine_impl(self):
        """Get the render engine for the text label using the current settings.

        Returns
        -------
            TextRenderEngine: The rendered engine.

        """
        selected_alignment = Direction(self.align.currentText())
        return TextRenderEngine(
            text_lines=self.label.toPlainText().splitlines(),
            font_file_name=self.font_style.currentData(),
            frame_width_px=self.frame_width_px.value(),
            font_size_ratio=self.font_size.value() / 100.0,
            align=selected_alignment,
        )


class QrDymoLabelWidget(BaseLabelWidget):
    """A widget for rendering QR codes on Dymo labels.

    Args:
    ----
        render_context (RenderContext): The render context to use for rendering
            the QR code.
        parent (QWidget, optional): The parent widget. Defaults to None.

    """

    def __init__(self, render_context, parent=None):
        """Initialize the QrDymoLabelWidget.

        Args:
        ----
            render_context (RenderContext): The render context to use for rendering
                the QR code.
            parent (QWidget, optional): The parent widget. Defaults to None.

        """
        super().__init__(parent)
        self.render_context = render_context

        self.label = QLineEdit("")
        layout = QHBoxLayout()
        item_icon = QLabel()
        item_icon.setPixmap(QIcon(str(ICON_DIR / "qr_icon.png")).pixmap(32, 32))
        item_icon.setAlignment(QtCore.Qt.AlignmentFlag.AlignHCenter)
        layout.addWidget(item_icon)
        layout.addWidget(self.label)
        self.label.textChanged.connect(self.content_changed)
        self.setLayout(layout)

    @property
    def render_engine_impl(self):
        """Get the render engine for the QR label using the current settings.

        Returns
        -------
            QrRenderEngine: The render engine.

        """
        try:
            return QrRenderEngine(content=self.label.text())
        except NoContentError:
            return EmptyRenderEngine()


class BarcodeDymoLabelWidget(BaseLabelWidget):
    """A widget for rendering barcode labels using the Dymo label printer.

    Args:
    ----
        render_context (RenderContext): An instance of the RenderContext class.
        parent (QWidget): The parent widget of this widget.

    Attributes:
    ----------
        render_context (RenderContext): An instance of the RenderContext class.
        label (QLineEdit): A QLineEdit widget for entering the content of the
            barcode label.
        Type (QComboBox): A QComboBox widget for selecting the type of barcode
            to render.
        font_style (FontStyle): The font style selection dropdown.
        font_size (QSpinBox): The font size selection spinner.
        frame_width_px (QSpinBox): The frame width selection spinner.
    Signals:
        content_changed(): Emitted when the content of the label or the selected
            barcode type changes.

    Methods:
    -------
        __init__(self, render_context, parent=None): Initializes the widget.
        render_label_impl(self): Renders the barcode label using the current content
            and barcode type.

    """

    label: QLineEdit
    barcode_type_label: QLabel
    barcode_type: QComboBox
    show_text_label: QLabel
    show_text_checkbox: QCheckBox
    font_style: FontStyle
    font_size: QSpinBox
    frame_width_px: QSpinBox
    font_label: QLabel
    size_label: QLabel
    frame_label: QLabel
    align_label: QLabel
    align: QComboBox

    def __init__(self, render_context, parent=None):
        super().__init__(parent)
        self.render_context = render_context

        self.label = QLineEdit("")

        # Hidable text fields and their labels
        self.font_label = QLabel("Font:")
        self.font_style = FontStyle()
        self.size_label = QLabel("Size [%]:")
        self.font_size = QSpinBox()
        self.font_size.setMaximum(150)
        self.font_size.setMinimum(0)
        self.font_size.setSingleStep(1)
        self.font_size.setValue(90)
        self.frame_label = QLabel("Frame Width:")
        self.frame_width_px = QSpinBox()
        self.align_label = QLabel("Alignment:")
        self.align = QComboBox()
        self.item_icon = QLabel()
        self.item_icon.setAlignment(QtCore.Qt.AlignmentFlag.AlignHCenter)

        self.align.addItems(["left", "center", "right"])
        # Set the default value to "center"
        self.align.setCurrentIndex(1)

        self.set_text_fields_visibility(True)

        layout = QHBoxLayout()

        self.barcode_type_label = QLabel("Type:")
        self.barcode_type = QComboBox()
        self.barcode_type.addItems(bt.value for bt in BarcodeType)

        # Checkbox for toggling text fields
        self.show_text_label = QLabel("Text:")
        self.show_text_checkbox = QCheckBox()
        self.show_text_checkbox.setChecked(True)
        self.show_text_checkbox.stateChanged.connect(
            self.toggle_text_fields_and_rerender
        )

        layout.addWidget(self.item_icon)
        layout.addWidget(self.label)
        layout.addWidget(self.barcode_type_label)
        layout.addWidget(self.barcode_type)
        layout.addWidget(self.show_text_label)
        layout.addWidget(self.show_text_checkbox)
        layout.addWidget(self.font_label)
        layout.addWidget(self.font_style)
        layout.addWidget(self.size_label)
        layout.addWidget(self.font_size)
        layout.addWidget(self.frame_label)
        layout.addWidget(self.frame_width_px)
        layout.addWidget(self.align_label)
        layout.addWidget(self.align)

        self.label.textChanged.connect(self.content_changed)
        self.frame_width_px.valueChanged.connect(self.content_changed)
        self.font_size.valueChanged.connect(self.content_changed)
        self.font_style.currentTextChanged.connect(self.content_changed)
        self.align.currentTextChanged.connect(self.content_changed)
        self.barcode_type.currentTextChanged.connect(self.content_changed)

        self.setLayout(layout)

    def set_text_fields_visibility(self, visible):
        self.font_label.setVisible(visible)
        self.font_style.setVisible(visible)
        self.size_label.setVisible(visible)
        self.font_size.setVisible(visible)
        self.frame_label.setVisible(visible)
        self.frame_width_px.setVisible(visible)
        self.align_label.setVisible(visible)
        self.align.setVisible(visible)
        if visible:
            self.item_icon.setPixmap(
                QIcon(str(ICON_DIR / "barcode_text_icon.png")).pixmap(32, 32)
            )
        else:
            self.item_icon.setPixmap(
                QIcon(str(ICON_DIR / "barcode_icon.png")).pixmap(32, 32)
            )

    def toggle_text_fields_and_rerender(self) -> None:
        is_checked = self.show_text_checkbox.isChecked()
        self.set_text_fields_visibility(is_checked)
        self.content_changed()  # Trigger rerender

    @property
    def render_engine_impl(self):
        """Get the render engine for the barcode label using the current settings.

        Returns
        -------
            RenderEngine: The rendered engine (either BarcodeRenderEngine or
            BarcodeWithTextRenderEngine).

        """
        render_engine: RenderEngine
        if self.show_text_checkbox.isChecked():
            render_engine = BarcodeWithTextRenderEngine(
                content=self.label.text(),
                barcode_type=self.barcode_type.currentText(),
                font_file_name=self.font_style.currentData(),
                frame_width_px=self.frame_width_px.value(),
                font_size_ratio=self.font_size.value() / 100.0,
                align=Direction(self.align.currentText()),
            )
        else:
            render_engine = BarcodeRenderEngine(
                content=self.label.text(),
                barcode_type=self.barcode_type.currentText(),
            )
        return render_engine


class ImageDymoLabelWidget(BaseLabelWidget):
    """A widget for rendering image-based Dymo labels.

    Args:
    ----
        context (RenderContext): The render context to use for rendering the label.
        parent (QWidget, optional): The parent widget. Defaults to None.

    """

    def __init__(self, render_context, parent=None):
        """Initialize the ImageDymoLabelWidget.

        Args:
        ----
            render_context (RenderContext): The render context to use for rendering
                the label.
            parent (QWidget, optional): The parent widget. Defaults to None.

        """
        super().__init__(parent)
        self.render_context = render_context

        self.label = QLineEdit("")
        layout = QHBoxLayout()
        item_icon = QLabel()
        item_icon.setPixmap(QIcon(str(ICON_DIR / "img_icon.png")).pixmap(32, 32))
        item_icon.setAlignment(QtCore.Qt.AlignmentFlag.AlignHCenter)

        button = QPushButton("Select file")
        file_dialog = QFileDialog()
        button.clicked.connect(
            lambda: self.label.setText(
                str(Path(file_dialog.getOpenFileName()[0]).absolute())
            )
        )

        layout.addWidget(item_icon)
        layout.addWidget(self.label)
        layout.addWidget(button)

        self.label.textChanged.connect(self.content_changed)
        self.setLayout(layout)

    @property
    def render_engine_impl(self):
        """Get the render engine for the image label using the current settings.

        Returns
        -------
            PictureRenderEngine: The rendered engine.

        """
        try:
            return PictureRenderEngine(picture_path=self.label.text())
        except PicturePathDoesNotExist:
            return EmptyRenderEngine()



from __future__ import annotations

import logging

from PyQt6 import QtCore
from PyQt6.QtCore import QTimer
from PyQt6.QtWidgets import QWidget
from usb.core import NoBackendError, USBError

from labelle.lib.devices.device_manager import DeviceManager, DeviceManagerError
from labelle.lib.devices.usb_device import UsbDevice

LOG = logging.getLogger(__name__)
POSSIBLE_USB_ERRORS = (NoBackendError, USBError)


class OnlineDeviceManager(QWidget):
    _last_scan_error: DeviceManagerError | None
    _status_time: QTimer
    _device_manager: DeviceManager
    last_scan_error_changed_signal = QtCore.pyqtSignal(
        name="lastScanErrorChangedSignal"
    )
    devices_changed_signal = QtCore.pyqtSignal(name="devicesChangedSignal")

    def __init__(self) -> None:
        super().__init__()
        self._device_manager = DeviceManager()
        self._last_scan_error = None
        self._init_timers()

    def _refresh_devices(self) -> None:
        prev = self._last_scan_error
        try:
            changed = self._device_manager.scan()
            self._last_scan_error = None
            if changed:
                self.devices_changed_signal.emit()
        except DeviceManagerError as e:
            self._last_scan_error = e

        if str(prev) != str(self._last_scan_error):
            self.devices_changed_signal.emit()
            self.last_scan_error_changed_signal.emit()

    def _init_timers(self) -> None:
        self._status_time = QTimer()
        self._status_time.timeout.connect(self._refresh_devices)
        self._status_time.start(2000)
        self._refresh_devices()

    @property
    def last_scan_error(self) -> DeviceManagerError | None:
        return self._last_scan_error

    @property
    def devices(self) -> list[UsbDevice]:
        return self._device_manager.devices



import logging
import traceback

from PyQt6.QtWidgets import QMessageBox, QWidget

from labelle.lib.logger import VERBOSE_NOTICE, is_verbose, print_exception

LOG = logging.getLogger(__name__)


def crash_msg_box(parent: QWidget, title: str, err: Exception) -> None:
    print_exception(err)
    text = f"{err!r}\n\n{traceback.format_exc() if is_verbose() else VERBOSE_NOTICE}"
    QMessageBox.warning(parent, title, text)



from pytestqt.qtbot import QtBot

from labelle.gui.gui import LabelleWindow
from labelle.gui.q_label_widgets import (
    TextDymoLabelWidget,
)


def test_main_window(qtbot: QtBot):
    widget = LabelleWindow()
    qtbot.addWidget(widget)

    assert not widget._actions.isEnabled()
    assert widget._device_selector.isEnabled()
    assert not widget._label_list.isEnabled()
    assert not widget._render_widget.isEnabled()
    assert not widget._render.isEnabled()
    assert not widget._settings_toolbar.isEnabled()
    assert widget._device_selector._error_label.text() == "No supported devices found"
    assert not widget._actions._print_button.isEnabled()
    assert widget._label_list.count() == 1
    item = widget._label_list.itemWidget(widget._label_list.item(0))
    assert isinstance(item, TextDymoLabelWidget)
    assert item.label.toPlainText() == "text"



import webbrowser
from tempfile import NamedTemporaryFile

import typer
from PIL import Image, ImageOps

from labelle.lib.constants import Output
from labelle.lib.unicode_blocks import image_to_unicode


def output_bitmap(bitmap: Image.Image, output: Output):
    if output in (Output.CONSOLE, Output.CONSOLE_INVERTED):
        label_rotated = bitmap.transpose(Image.Transpose.ROTATE_270)
        invert = output == Output.CONSOLE_INVERTED
        typer.echo(image_to_unicode(label_rotated, invert=invert))
    if output == Output.IMAGEMAGICK:
        ImageOps.invert(bitmap.convert("RGB")).show()
    if output == Output.BROWSER:
        with NamedTemporaryFile(suffix=".png", delete=False) as fp:
            inverted = ImageOps.invert(bitmap.convert("RGB"))
            ImageOps.invert(inverted).save(fp)
            webbrowser.open(f"file://{fp.name}")
    if output == Output.PNG:
        bitmap.save("output.png")
        typer.echo("Saved output.png")



from PIL import Image

from labelle.lib.constants import QRCode
from labelle.lib.render_engines import NoContentError
from labelle.lib.render_engines.render_context import RenderContext
from labelle.lib.render_engines.render_engine import (
    RenderEngine,
    RenderEngineException,
)
from labelle.lib.utils import draw_image, scaling


class QrTooBigError(RenderEngineException):
    def __init__(self) -> None:
        msg = "Too much information to store in the QR code"
        super().__init__(msg)


class QrRenderEngine(RenderEngine):
    _content: str

    def __init__(self, content: str):
        super().__init__()
        if not len(content):
            raise NoContentError()
        self._content = content

    def render(self, context: RenderContext) -> Image.Image:
        # create QR object from first string
        code = QRCode(self._content, error="M")
        qr_text_lines = code.text(quiet_zone=1).split()

        # create an empty label image
        height_px = context.height_px
        qr_scale = height_px // len(qr_text_lines)
        qr_offset = (height_px - len(qr_text_lines) * qr_scale) // 2
        label_width_px = len(qr_text_lines[0]) * qr_scale

        if not qr_scale:
            raise QrTooBigError()

        bitmap = Image.new("1", (label_width_px, height_px))

        with draw_image(bitmap) as draw:
            # write the qr-code into the empty image
            for i, line in enumerate(qr_text_lines):
                for j, char in enumerate(line):
                    if char == "1":
                        qr_pixels = scaling(
                            (j * qr_scale, i * qr_scale + qr_offset), qr_scale
                        )
                        draw.point(qr_pixels, 1)
        return bitmap



"""
A module for finding, managing, and using fonts across platforms.

This module provides a single `FontManager` instance, ``fontManager``, that can
be shared across backends and platforms.  The `findfont`
function returns the best TrueType (TTF) font file in the local or
system font path that matches the specified `FontProperties`
instance.  The `FontManager` also handles Adobe Font Metrics
(AFM) font files for use by the PostScript backend.

The design is based on the `W3C Cascading Style Sheet, Level 1 (CSS1)
font specification <http://www.w3.org/TR/1998/REC-CSS2-19980512/>`_.
Future versions may implement the Level 2 or 2.1 specifications.
"""

# KNOWN ISSUES
#
#   - documentation
#   - font variant is untested
#   - font stretch is incomplete
#   - font size is incomplete
#   - default font algorithm needs improvement and testing
#   - setWeights function needs improvement
#   - 'light' is an invalid weight value, remove it.

from collections import namedtuple
from functools import lru_cache
import logging
import os
from pathlib import Path
import subprocess
import sys


_log = logging.getLogger(__name__)

font_scalings = {
    'xx-small': 0.579,
    'x-small':  0.694,
    'small':    0.833,
    'medium':   1.0,
    'large':    1.200,
    'x-large':  1.440,
    'xx-large': 1.728,
    'larger':   1.2,
    'smaller':  0.833,
    None:       1.0,
}
stretch_dict = {
    'ultra-condensed': 100,
    'extra-condensed': 200,
    'condensed':       300,
    'semi-condensed':  400,
    'normal':          500,
    'semi-expanded':   600,
    'semi-extended':   600,
    'expanded':        700,
    'extended':        700,
    'extra-expanded':  800,
    'extra-extended':  800,
    'ultra-expanded':  900,
    'ultra-extended':  900,
}
weight_dict = {
    'ultralight': 100,
    'light':      200,
    'normal':     400,
    'regular':    400,
    'book':       400,
    'medium':     500,
    'roman':      500,
    'semibold':   600,
    'demibold':   600,
    'demi':       600,
    'bold':       700,
    'heavy':      800,
    'extra bold': 800,
    'black':      900,
}
_weight_regexes = [
    # From fontconfig's FcFreeTypeQueryFaceInternal; not the same as
    # weight_dict!
    ("thin", 100),
    ("extralight", 200),
    ("ultralight", 200),
    ("demilight", 350),
    ("semilight", 350),
    ("light", 300),  # Needs to come *after* demi/semilight!
    ("book", 380),
    ("regular", 400),
    ("normal", 400),
    ("medium", 500),
    ("demibold", 600),
    ("demi", 600),
    ("semibold", 600),
    ("extrabold", 800),
    ("superbold", 800),
    ("ultrabold", 800),
    ("bold", 700),  # Needs to come *after* extra/super/ultrabold!
    ("ultrablack", 1000),
    ("superblack", 1000),
    ("extrablack", 1000),
    (r"\bultra", 1000),
    ("black", 900),  # Needs to come *after* ultra/super/extrablack!
    ("heavy", 900),
]
font_family_aliases = {
    'serif',
    'sans-serif',
    'sans serif',
    'cursive',
    'fantasy',
    'monospace',
    'sans',
}

_ExceptionProxy = namedtuple('_ExceptionProxy', ['klass', 'message'])

# OS Font paths
try:
    _HOME = Path.home()
except Exception:  # Exceptions thrown by home() are not specified...
    _HOME = Path(os.devnull)  # Just an arbitrary path with no children.
MSFolders = \
    r'Software\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders'
MSFontDirectories = [
    r'SOFTWARE\Microsoft\Windows NT\CurrentVersion\Fonts',
    r'SOFTWARE\Microsoft\Windows\CurrentVersion\Fonts']
MSUserFontDirectories = [
    str(_HOME / 'AppData/Local/Microsoft/Windows/Fonts'),
    str(_HOME / 'AppData/Roaming/Microsoft/Windows/Fonts'),
]
X11FontDirectories = [
    # an old standard installation point
    "/usr/X11R6/lib/X11/fonts/TTF/",
    "/usr/X11/lib/X11/fonts",
    # here is the new standard location for fonts
    "/usr/share/fonts/",
    # documented as a good place to install new fonts
    "/usr/local/share/fonts/",
    # common application, not really useful
    "/usr/lib/openoffice/share/fonts/truetype/",
    # user fonts
    str((Path(os.environ.get('XDG_DATA_HOME') or _HOME / ".local/share"))
        / "fonts"),
    str(_HOME / ".fonts"),
]
OSXFontDirectories = [
    "/Library/Fonts/",
    "/Network/Library/Fonts/",
    "/System/Library/Fonts/",
    # fonts installed via MacPorts
    "/opt/local/share/fonts",
    # user fonts
    str(_HOME / "Library/Fonts"),
]


def get_fontext_synonyms(fontext):
    """
    Return a list of file extensions that are synonyms for
    the given file extension *fileext*.
    """
    return {
        'afm': ['afm'],
        'otf': ['otf', 'ttc', 'ttf'],
        'ttc': ['otf', 'ttc', 'ttf'],
        'ttf': ['otf', 'ttc', 'ttf'],
    }[fontext]


def list_fonts(directory, extensions):
    """
    Return a list of all fonts matching any of the extensions, found
    recursively under the directory.
    """
    extensions = ["." + ext for ext in extensions]
    return [os.path.join(dirpath, filename)
            # os.walk ignores access errors, unlike Path.glob.
            for dirpath, _, filenames in os.walk(directory)
            for filename in filenames
            if Path(filename).suffix.lower() in extensions]


def win32FontDirectory():
    r"""
    Return the user-specified font directory for Win32.  This is
    looked up from the registry key ::

      \\HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders\Fonts

    If the key is not found, ``%WINDIR%\Fonts`` will be returned.
    """
    import winreg
    try:
        with winreg.OpenKey(winreg.HKEY_CURRENT_USER, MSFolders) as user:
            return winreg.QueryValueEx(user, 'Fonts')[0]
    except OSError:
        return os.path.join(os.environ['WINDIR'], 'Fonts')


def _get_win32_installed_fonts():
    """List the font paths known to the Windows registry."""
    import winreg
    items = set()
    # Search and resolve fonts listed in the registry.
    for domain, base_dirs in [
            (winreg.HKEY_LOCAL_MACHINE, [win32FontDirectory()]),  # System.
            (winreg.HKEY_CURRENT_USER, MSUserFontDirectories),  # User.
    ]:
        for base_dir in base_dirs:
            for reg_path in MSFontDirectories:
                try:
                    with winreg.OpenKey(domain, reg_path) as local:
                        for j in range(winreg.QueryInfoKey(local)[1]):
                            # value may contain the filename of the font or its
                            # absolute path.
                            key, value, tp = winreg.EnumValue(local, j)
                            if not isinstance(value, str):
                                continue
                            try:
                                # If value contains already an absolute path,
                                # then it is not changed further.
                                path = Path(base_dir, value).resolve()
                            except RuntimeError:
                                # Don't fail with invalid entries.
                                continue
                            items.add(path)
                except (OSError, MemoryError):
                    continue
    return items


@lru_cache()
def _get_fontconfig_fonts():
    """Cache and list the font paths known to ``fc-list``."""
    try:
        if b'--format' not in subprocess.check_output(['fc-list', '--help']):
            _log.warning(  # fontconfig 2.7 implemented --format.
                'Matplotlib needs fontconfig>=2.7 to query system fonts.')
            return []
        out = subprocess.check_output(['fc-list', '--format=%{file}\\n'])
    except (OSError, subprocess.CalledProcessError):
        return []
    return [Path(os.fsdecode(fname)) for fname in out.split(b'\n')]


def findSystemFonts(fontpaths=None, fontext='ttf'):
    """
    Search for fonts in the specified font paths.  If no paths are
    given, will use a standard set of system paths, as well as the
    list of fonts tracked by fontconfig if fontconfig is installed and
    available.  A list of TrueType fonts are returned by default with
    AFM fonts as an option.
    """
    fontfiles = set()
    fontexts = get_fontext_synonyms(fontext)

    if fontpaths is None:
        if sys.platform == 'win32':
            installed_fonts = _get_win32_installed_fonts()
            fontpaths = []
        else:
            installed_fonts = _get_fontconfig_fonts()
            if sys.platform == 'darwin':
                fontpaths = [*X11FontDirectories, *OSXFontDirectories]
            else:
                fontpaths = X11FontDirectories
        fontfiles.update(str(path) for path in installed_fonts
                         if path.suffix.lower()[1:] in fontexts)

    elif isinstance(fontpaths, str):
        fontpaths = [fontpaths]

    for path in fontpaths:
        fontfiles.update(map(os.path.abspath, list_fonts(path, fontexts)))

    return [fname for fname in fontfiles if os.path.exists(fname)]



from __future__ import annotations

import logging
from dataclasses import dataclass

from PyQt6 import QtCore
from PyQt6.QtWidgets import QCheckBox, QComboBox, QLabel, QSpinBox, QToolBar, QWidget

from labelle.lib.constants import Direction

LOG = logging.getLogger(__name__)


FOREGROUND_COLOR__VALUES = ["black", "white", "yellow", "blue", "red", "green"]
BACKGROUND_COLOR__VALUES = ["white", "black", "yellow", "blue", "red", "green"]
HORIZONTAL_MARGIN_MM__MAX_VALUE = 100
MIN_LABEL_WIDTH_MM__MIN_VALUE = 300
PREVIEW_SHOW_MARGINS__DEFAULT_VALUE = False


@dataclass
class Settings:
    background_color: str
    foreground_color: str
    horizontal_margin_mm: float
    justify: Direction
    min_label_width_mm: float
    preview_show_margins: bool
    tape_size_mm: int


class QSettingsToolbar(QToolBar):
    settings_changed_signal = QtCore.pyqtSignal(Settings, name="settingsChangedSignal")

    def __init__(self, parent: QWidget | None = None):
        super().__init__(parent)
        self._background_color = QComboBox()
        self._foreground_color = QComboBox()
        self._horizontal_margin_mm = QSpinBox()
        self._justify = QComboBox()
        self._min_label_width_mm = QSpinBox()
        self._preview_show_margins = QCheckBox()
        self._tape_size_mm = QComboBox()

        self._init_elements()
        self._init_connections()
        self._init_layout()
        self.on_settings_changed()

    def _init_elements(self) -> None:
        self._horizontal_margin_mm.setMaximum(HORIZONTAL_MARGIN_MM__MAX_VALUE)
        self._min_label_width_mm.setMaximum(MIN_LABEL_WIDTH_MM__MIN_VALUE)
        self._justify.addItems(d.value for d in Direction)
        self._foreground_color.addItems(FOREGROUND_COLOR__VALUES)
        self._background_color.addItems(BACKGROUND_COLOR__VALUES)
        self._preview_show_margins.setChecked(PREVIEW_SHOW_MARGINS__DEFAULT_VALUE)

    def update_labeler_context(
        self,
        supported_tape_sizes: tuple[int, ...],
        installed_tape_size: int,
        minimum_horizontal_margin_mm: float,
    ) -> None:
        for tape_size_mm in supported_tape_sizes:
            self._tape_size_mm.addItem(str(tape_size_mm), tape_size_mm)
        tape_size_index = supported_tape_sizes.index(installed_tape_size)
        self._tape_size_mm.setCurrentIndex(tape_size_index)

        h_margins_mm = round(minimum_horizontal_margin_mm)
        self._horizontal_margin_mm.setMinimum(h_margins_mm)
        if not self._horizontal_margin_mm.value():
            self._horizontal_margin_mm.setValue(h_margins_mm)
        self._min_label_width_mm.setMinimum(h_margins_mm * 2)

    def _init_connections(self) -> None:
        self._background_color.currentTextChanged.connect(self.on_settings_changed)
        self._foreground_color.currentTextChanged.connect(self.on_settings_changed)
        self._horizontal_margin_mm.valueChanged.connect(self.on_settings_changed)
        self._justify.currentTextChanged.connect(self.on_settings_changed)
        self._min_label_width_mm.valueChanged.connect(self.on_settings_changed)
        self._preview_show_margins.stateChanged.connect(self.on_settings_changed)
        self._tape_size_mm.currentTextChanged.connect(self.on_settings_changed)

    def _init_layout(self) -> None:
        self.addWidget(QLabel("Margin [mm]:"))
        self.addWidget(self._horizontal_margin_mm)
        self.addSeparator()
        self.addWidget(QLabel("Tape Size [mm]:"))
        self.addWidget(self._tape_size_mm)
        self.addSeparator()
        self.addWidget(QLabel("Min Label Length [mm]:"))
        self.addWidget(self._min_label_width_mm)
        self.addSeparator()
        self.addWidget(QLabel("Justify:"))
        self.addWidget(self._justify)
        self.addSeparator()
        self.addWidget(QLabel("Tape Colors: "))
        self.addWidget(self._foreground_color)
        self.addWidget(QLabel(" on "))
        self.addWidget(self._background_color)
        self.addWidget(QLabel("Show margins:"))
        self.addWidget(self._preview_show_margins)

    @property
    def settings(self) -> Settings:
        return Settings(
            background_color=self._background_color.currentText(),
            foreground_color=self._foreground_color.currentText(),
            horizontal_margin_mm=self._horizontal_margin_mm.value(),
            justify=Direction(self._justify.currentText()),
            min_label_width_mm=self._min_label_width_mm.value(),
            preview_show_margins=self._preview_show_margins.isChecked(),
            tape_size_mm=self._tape_size_mm.currentData(),
        )

    def on_settings_changed(self) -> None:
        self.settings_changed_signal.emit(self.settings)



# === LICENSE STATEMENT ===
# Copyright (c) 2011 Sebastian J. Bronner <waschtl@sbronner.com>
#
# Copying and distribution of this file, with or without modification, are
# permitted in any medium without royalty provided the copyright notice and
# this notice are preserved.
# === END LICENSE STATEMENT ===
import logging
import sys
from pathlib import Path
from typing import List, NoReturn, Optional

import typer
from rich.console import Console
from rich.table import Table

if sys.version_info >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated

from labelle import __version__
from labelle.lib.constants import (
    DEFAULT_BARCODE_TYPE,
    DEFAULT_MARGIN_PX,
    PIXELS_PER_MM,
    USE_QR,
    BarcodeType,
    Direction,
    Output,
    e_qrcode,
)
from labelle.lib.devices.device_manager import DeviceManager, DeviceManagerNoDevices
from labelle.lib.devices.dymo_labeler import DymoLabeler
from labelle.lib.env_config import is_verbose_env_vars
from labelle.lib.font_config import (
    DefaultFontStyle,
    FontStyle,
    NoFontFound,
    get_available_fonts,
    get_font_path,
)
from labelle.lib.logger import configure_logging, set_not_verbose
from labelle.lib.outputs import output_bitmap
from labelle.lib.render_engines import (
    BarcodeRenderEngine,
    BarcodeWithTextRenderEngine,
    HorizontallyCombinedRenderEngine,
    PictureRenderEngine,
    PrintPayloadRenderEngine,
    PrintPreviewRenderEngine,
    QrRenderEngine,
    RenderContext,
    RenderEngine,
    SamplePatternRenderEngine,
    TextRenderEngine,
)

LOG = logging.getLogger(__name__)


def mm_to_payload_px(mm: float, margin: float) -> float:
    """Convert a length in mm to a number of pixels of payload.

    The print resolution is 7 pixels/mm, and margin is subtracted from each side.
    """
    return max(0, (mm * PIXELS_PER_MM) - margin * 2)


def version_callback(value: bool) -> None:
    if value:
        typer.echo(f"Labelle: {__version__}")
        raise typer.Exit()


def qr_callback(qr_content: str) -> str:
    # check if barcode, qrcode or text should be printed, use frames only on text
    if qr_content and not USE_QR:
        raise typer.BadParameter(
            "QR code cannot be used without QR support installed"
        ) from e_qrcode
    return qr_content


def get_device_manager() -> DeviceManager:
    device_manager = DeviceManager()
    try:
        device_manager.scan()
    except DeviceManagerNoDevices as e:
        err_console = Console(stderr=True)
        err_console.print(f"Error: {e}")
        raise typer.Exit() from e
    return device_manager


app = typer.Typer()


@app.command(hidden=True)
def list_devices() -> NoReturn:
    device_manager = get_device_manager()
    console = Console()
    headers = ["Manufacturer", "Product", "Serial Number", "USB"]
    table = Table(*headers, show_header=True)
    for device in device_manager.devices:
        table.add_row(
            device.manufacturer, device.product, device.serial_number, device.usb_id
        )
    console.print(table)
    raise typer.Exit()


@app.callback(invoke_without_command=True)
def default(
    ctx: typer.Context,
    version: Annotated[
        Optional[bool],
        typer.Option(
            "--version",
            callback=version_callback,
            is_eager=True,
            help="Show application version",
        ),
    ] = None,
    device_pattern: Annotated[
        Optional[List[str]],
        typer.Option(
            "--device",
            help=(
                "Select a particular device by filtering for a given substring "
                "in the device's manufacturer, product or serial number"
            ),
            rich_help_panel="Device Configuration",
        ),
    ] = None,
    text: Annotated[
        Optional[List[str]],
        typer.Argument(
            help="Text, each parameter gives a new line",
            rich_help_panel="Elements",
        ),
    ] = None,
    verbose: Annotated[
        bool, typer.Option("--verbose", "-v", help="Increase logging verbosity")
    ] = False,
    style: Annotated[
        FontStyle, typer.Option(help="Set fonts style", rich_help_panel="Design")
    ] = DefaultFontStyle,
    frame_width_px: Annotated[
        int,
        typer.Option(
            help="Draw frame of given width [px] around text",
            rich_help_panel="Design",
        ),
    ] = 0,
    align: Annotated[
        Direction, typer.Option(help="Align multiline text", rich_help_panel="Design")
    ] = Direction.LEFT,
    justify: Annotated[
        Direction,
        typer.Option(
            help="Justify content of label if label content is less than the minimum or"
            " fixed length",
            rich_help_panel="Design",
        ),
    ] = Direction.LEFT,
    sample_pattern: Annotated[
        Optional[int],
        typer.Option(help="Prints test pattern of a desired dot height [px]"),
    ] = None,
    min_length: Annotated[
        Optional[float],
        typer.Option(
            help="Minimum label length [mm], add whitespace if smaller",
            rich_help_panel="Label Dimensions",
        ),
    ] = None,
    max_length: Annotated[
        Optional[float],
        typer.Option(
            help="Maximum label length [mm], error if the label won't fit",
            rich_help_panel="Label Dimensions",
        ),
    ] = None,
    fixed_length: Annotated[
        Optional[float],
        typer.Option(
            help="Fixed label length [mm], set both minimum and maximum length",
            rich_help_panel="Label Dimensions",
        ),
    ] = None,
    output: Annotated[
        Output,
        typer.Option(help="Destination of the label render"),
    ] = Output.PRINTER,
    font: Annotated[
        Optional[str],
        typer.Option(
            help="User font. Overrides --style parameter", rich_help_panel="Design"
        ),
    ] = None,
    qr_content: Annotated[
        Optional[str],
        typer.Option(
            "--qr", callback=qr_callback, help="QR code", rich_help_panel="Elements"
        ),
    ] = None,
    batch: Annotated[
        bool,
        typer.Option(help="Read batch commands from stdin", rich_help_panel="Elements"),
    ] = False,
    barcode_content: Annotated[
        Optional[str],
        typer.Option("--barcode", help="Barcode", rich_help_panel="Elements"),
    ] = None,
    barcode_type: Annotated[
        BarcodeType,
        typer.Option(
            help="Barcode type",
            rich_help_panel="Elements",
        ),
    ] = DEFAULT_BARCODE_TYPE,
    barcode_with_text_content: Annotated[
        Optional[str],
        typer.Option(
            "--barcode-with-text", help="Barcode with text", rich_help_panel="Elements"
        ),
    ] = None,
    picture: Annotated[
        Optional[Path], typer.Option(help="Picture", rich_help_panel="Elements")
    ] = None,
    margin_px: Annotated[
        float,
        typer.Option(
            help="Horizontal margins [px]", rich_help_panel="Label Dimensions"
        ),
    ] = DEFAULT_MARGIN_PX,
    font_scale: Annotated[
        float,
        typer.Option(help="Scaling font factor, [0,100] [%]", rich_help_panel="Design"),
    ] = 90,
    tape_size_mm: Annotated[
        Optional[int],
        typer.Option(help="Tape size [mm]", rich_help_panel="Device Configuration"),
    ] = None,
    # Old dymoprint arguments
    preview: Annotated[
        bool,
        typer.Option(
            "--preview",
            "-n",
            help="DEPRECATED",
            hidden=True,
        ),
    ] = False,
    preview_inverted: Annotated[
        bool,
        typer.Option(
            "--preview-inverted",
            help="DEPRECATED",
            hidden=True,
        ),
    ] = False,
    imagemagick: Annotated[
        bool,
        typer.Option(
            "--imagemagick",
            help="DEPRECATED",
            hidden=True,
        ),
    ] = False,
    old_help: Annotated[
        bool,
        typer.Option(
            "-h",
            help="DEPRECATED",
            hidden=True,
        ),
    ] = False,
    old_frame: Annotated[
        bool,
        typer.Option(
            "-f",
            help="DEPRECATED",
            hidden=True,
        ),
    ] = False,
    old_style: Annotated[
        Optional[str],
        typer.Option(
            "-s",
            help="DEPRECATED",
            hidden=True,
        ),
    ] = None,
    old_align: Annotated[
        Optional[str],
        typer.Option(
            "-a",
            help="DEPRECATED",
            hidden=True,
        ),
    ] = None,
    old_font: Annotated[
        Optional[str],
        typer.Option(
            "-u",
            help="DEPRECATED",
            hidden=True,
        ),
    ] = None,
    old_barcode: Annotated[
        Optional[str],
        typer.Option(
            "-c",
            help="DEPRECATED",
            hidden=True,
        ),
    ] = None,
    barcode_text: Annotated[
        Optional[str],
        typer.Option(
            "--barcode-text",
            help="DEPRECATED",
            hidden=True,
        ),
    ] = None,
    old_picture: Annotated[
        Optional[str],
        typer.Option(
            "-p",
            help="DEPRECATED",
            hidden=True,
        ),
    ] = None,
    old_margin: Annotated[
        Optional[int],
        typer.Option(
            "-m",
            help="DEPRECATED",
            hidden=True,
        ),
    ] = None,
    scale: Annotated[
        Optional[float],
        typer.Option(
            help="DEPRECATED",
            hidden=True,
        ),
    ] = None,
    old_tape_size: Annotated[
        Optional[int],
        typer.Option(
            "-t",
            help="DEPRECATED",
            hidden=True,
        ),
    ] = None,
    old_min_length: Annotated[
        Optional[float],
        typer.Option(
            "-l",
            help="DEPRECATED",
            hidden=True,
        ),
    ] = None,
    old_justify: Annotated[
        Optional[str],
        typer.Option(
            "-j",
            help="DEPRECATED",
            hidden=True,
        ),
    ] = None,
    test_pattern: Annotated[
        Optional[int],
        typer.Option(
            help="DEPRECATED",
            hidden=True,
        ),
    ] = None,
) -> None:
    if ctx.invoked_subcommand is not None:
        return

    if (not verbose) and (not is_verbose_env_vars()):
        # Neither --verbose flag nor the environment variable is set.
        set_not_verbose()

    # Raise informative errors with old dymoprint arguments
    if preview:
        raise typer.BadParameter(
            "The --preview / -n flag is deprecated. Use --output=console instead."
        )
    if preview_inverted:
        raise typer.BadParameter(
            "The --preview-inverted flag is deprecated. Use "
            "--output=console-inverted instead."
        )
    if imagemagick:
        raise typer.BadParameter(
            "The --imagemagick flag is deprecated. Use --output=imagemagick instead."
        )
    if old_help:
        raise typer.BadParameter("The -h flag is deprecated. Use --help instead.")
    if old_frame:
        raise typer.BadParameter(
            "The -f flag is deprecated. Use --frame-width-px instead."
        )
    if old_style is not None:
        raise typer.BadParameter("The -s flag is deprecated. Use --style instead.")
    if old_align is not None:
        raise typer.BadParameter("The -a flag is deprecated. Use --align instead.")
    if old_font is not None:
        raise typer.BadParameter("The -u flag is deprecated. Use --font instead.")
    if old_barcode is not None:
        raise typer.BadParameter("The -c flag is deprecated. Use --barcode instead.")
    if barcode_text is not None:
        raise typer.BadParameter(
            "The --barcode-text flag is deprecated. Use --barcode-with-text instead."
        )
    if old_picture is not None:
        raise typer.BadParameter("The -p flag is deprecated. Use --picture instead.")
    if old_margin is not None:
        raise typer.BadParameter("The -m flag is deprecated. Use --margin-px instead.")
    if scale is not None:
        raise typer.BadParameter(
            "The --scale flag is deprecated. Use --font-scale instead."
        )
    if old_tape_size is not None:
        raise typer.BadParameter(
            "The -t flag is deprecated. Use --tape-size-mm instead."
        )
    if old_min_length is not None:
        raise typer.BadParameter("The -l flag is deprecated. Use --min-length instead.")
    if old_justify is not None:
        raise typer.BadParameter("The -j flag is deprecated. Use --justify instead.")
    if test_pattern is not None:
        raise typer.BadParameter(
            "The --test-pattern flag is deprecated. Use --sample-pattern instead."
        )

    # read config file
    try:
        font_path = get_font_path(font=font, style=style)
    except NoFontFound as e:
        valid_font_names = [f.stem for f in get_available_fonts()]
        msg = f"{e}. Valid fonts are: {', '.join(valid_font_names)}"
        raise typer.BadParameter(msg) from None

    if barcode_with_text_content and barcode_content:
        raise typer.BadParameter(
            "Cannot specify both barcode with text and regular barcode"
        )

    if fixed_length is not None and (min_length is not None or max_length is not None):
        raise typer.BadParameter(
            "Cannot specify min/max and fixed length at the same time"
        )

    if min_length is None:
        min_length = 0.0
    if min_length < 0:
        raise typer.BadParameter("Minimum length must be non-negative number")
    if max_length is not None:
        if max_length <= 0:
            raise typer.BadParameter("Maximum length must be positive number")
        if max_length < min_length:
            raise typer.BadParameter("Maximum length is less than minimum length")

    render_engines: list[RenderEngine] = []

    if sample_pattern:
        render_engines.append(SamplePatternRenderEngine(sample_pattern))

    if qr_content:
        render_engines.append(QrRenderEngine(qr_content))

    if barcode_with_text_content:
        render_engines.append(
            BarcodeWithTextRenderEngine(
                content=barcode_with_text_content,
                barcode_type=barcode_type,
                font_file_name=font_path,
                frame_width_px=frame_width_px,
            )
        )

    if barcode_content:
        render_engines.append(
            BarcodeRenderEngine(content=barcode_content, barcode_type=barcode_type)
        )

    def render_text(lines):
        render_engines.append(
            TextRenderEngine(
                text_lines=lines,
                font_file_name=font_path,
                frame_width_px=frame_width_px,
                font_size_ratio=int(font_scale) / 100.0,
                align=align,
            )
        )

    if text:
        render_text(text)

    if picture:
        render_engines.append(PictureRenderEngine(picture))

    if batch:
        accumulator: List[str] = []
        accumulator_type: str = "empty"
        accumulator_options: List[str] = []

        def flush_all():
            nonlocal accumulator
            nonlocal accumulator_type
            nonlocal accumulator_options
            if accumulator_type == "text":
                render_text(accumulator)
            elif accumulator_type == "qr":
                render_engines.append(
                    QrRenderEngine(qr_callback("\n".join(accumulator)))
                )
            elif accumulator_type == "barcode":
                barcode_type = (
                    BarcodeType(accumulator_options[0].lower())
                    if accumulator_options
                    else DEFAULT_BARCODE_TYPE
                )
                render_engines.append(
                    BarcodeRenderEngine("\n".join(accumulator), barcode_type)
                )

            accumulator = []
            accumulator_type = "empty"
            accumulator_options = []

        # Verify version
        line: str = sys.stdin.readline().strip()
        parts = line.split(":", 1)
        if not (parts[0] == "LABELLE-LABEL-SPEC-VERSION" and parts[1] == "1"):
            err_console = Console(stderr=True)
            err_console.print(
                "Error: Batch doesn't begin with LABELLE-LABEL-SPEC-VERSION:1"
            )
            raise typer.Exit()

        for line in sys.stdin:
            line = line.rstrip("\r\n")
            settings, value = line.split(":", 1)
            setting, *options = settings.split("#", 1)

            if setting == "TEXT":
                flush_all()
                accumulator_type = "text"
                accumulator.append(value)
            elif setting == "QR":
                flush_all()
                accumulator_type = "qr"
                accumulator.append(value)
            elif setting == "BARCODE":
                flush_all()
                accumulator_type = "barcode"
                accumulator_options = options
                accumulator.append(value)
            elif setting == "NEWLINE":
                accumulator.append(value)
            else:
                print("WARNING: invalid command", line)
        flush_all()

    if fixed_length is None:
        min_label_mm_len = min_length
        max_label_mm_len = max_length
    else:
        min_label_mm_len = fixed_length
        max_label_mm_len = fixed_length

    min_payload_len_px = mm_to_payload_px(min_label_mm_len, margin_px)
    max_payload_len_px = (
        mm_to_payload_px(max_label_mm_len, margin_px)
        if max_label_mm_len is not None
        else None
    )

    if output == Output.PRINTER:
        device_manager = get_device_manager()
        device = device_manager.find_and_select_device(patterns=device_pattern)
        device.setup()
    else:
        device = None

    dymo_labeler = DymoLabeler(tape_size_mm=tape_size_mm, device=device)
    if not render_engines:
        raise typer.BadParameter("No elements to print")
    render_engine = HorizontallyCombinedRenderEngine(render_engines)
    render_context = RenderContext(
        background_color="white",
        foreground_color="black",
        height_px=dymo_labeler.height_px,
        preview_show_margins=False,
    )

    # print or show the label
    render: RenderEngine
    if output == Output.PRINTER:
        render = PrintPayloadRenderEngine(
            render_engine=render_engine,
            justify=justify,
            visible_horizontal_margin_px=margin_px,
            labeler_margin_px=dymo_labeler.labeler_margin_px,
            max_width_px=max_payload_len_px,
            min_width_px=min_payload_len_px,
        )
        bitmap, _ = render.render_with_meta(render_context)
        dymo_labeler.print(bitmap)
    else:
        render = PrintPreviewRenderEngine(
            render_engine=render_engine,
            justify=justify,
            visible_horizontal_margin_px=margin_px,
            labeler_margin_px=dymo_labeler.labeler_margin_px,
            max_width_px=max_payload_len_px,
            min_width_px=min_payload_len_px,
        )
        bitmap = render.render(render_context)
        output_bitmap(bitmap, output)


def main() -> None:
    configure_logging()
    app()


if __name__ == "__main__":
    main()



import logging
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional

import labelle.resources.fonts
from labelle._vendor.matplotlib import font_manager
from labelle.lib.config_file import get_config_section

logger = logging.getLogger(__name__)


class NoFontFound(ValueError):
    def __init__(self, name):
        msg = f"No font named {name} found"
        super().__init__(msg)


class NoStyleFound(ValueError):
    def __init__(self, style):
        msg = f"No style named {style} found"
        super().__init__(msg)


_DEFAULT_FONTS_DIR = Path(labelle.resources.fonts.__file__).parent
_DEFAULT_STYLE = "regular"
_DEFAULT_STYLES_TO_FONT_PATH: Dict[str, Path] = {
    "regular": _DEFAULT_FONTS_DIR / "Carlito-Regular.ttf",
    "bold": _DEFAULT_FONTS_DIR / "Carlito-Bold.ttf",
    "italic": _DEFAULT_FONTS_DIR / "Carlito-Italic.ttf",
    "narrow": _DEFAULT_FONTS_DIR / "Carlito-BoldItalic.ttf",
}


class FontStyle(str, Enum):
    REGULAR = "regular"
    BOLD = "bold"
    ITALIC = "italic"
    NARROW = "narrow"


DefaultFontStyle = FontStyle(_DEFAULT_STYLE)


def _get_styles_to_font_path_lookup() -> Dict[str, Path]:
    """Get a lookup table for styles to font paths.

    The lookup table is read from the config file, if available.
    """
    styles_to_font_path = _DEFAULT_STYLES_TO_FONT_PATH.copy()
    fonts_config = get_config_section("FONTS")
    if fonts_config is not None:
        for style_from_config, filename in fonts_config.items():
            styles_to_font_path[style_from_config] = Path(filename)
    return styles_to_font_path


def get_font_path(
    font: Optional[str] = None, style: Optional[str] = _DEFAULT_STYLE
) -> Path:
    """Get the path to a font.

    The `font` argument can be either a font name or a path to a font file.
    If `font` is not provided, the default font is used. In that case, the `style`
    argument can be used to specify the style of the default font.
    """
    if font is not None:
        if Path(font).is_file():
            path = Path(font)
        else:
            path = _path_from_name(name=font)
    else:
        styles_to_font_path = _get_styles_to_font_path_lookup()
        if style in styles_to_font_path:
            path = styles_to_font_path[style]
        else:
            logger.debug(f"Style '{style}' unrecognized. Known: {styles_to_font_path}")
            raise NoStyleFound(style)

    # Double-check that the file exists
    if not Path(path).is_file():
        logger.error(f"Font file not found: {path}")
        raise NoFontFound(font)
    return path


def _path_from_name(name: str) -> Path:
    """Get the path to a font from its name.

    The name should be the name of the font file, without the extension.
    It is case-insensitive.
    """
    available_fonts = get_available_fonts()
    matching_fonts = [f for f in available_fonts if name.lower() == f.stem.lower()]
    if len(matching_fonts) == 0:
        raise NoFontFound(name)
    return matching_fonts[0]


def get_available_fonts() -> List[Path]:
    """Get a list of available font files."""
    fonts = [f for f in _DEFAULT_FONTS_DIR.iterdir() if f.suffix == ".ttf"]
    fonts.extend(Path(f) for f in font_manager.findSystemFonts())
    return sorted(fonts, key=lambda f: f.stem.lower())



import logging
import sys

_IS_VERBOSE = True
LOG = logging.getLogger("labelle")
VERBOSE_NOTICE = "Run with --verbose for more information"


def _update_log_level() -> None:
    LOG.setLevel(logging.DEBUG if _IS_VERBOSE else logging.INFO)


def set_not_verbose() -> None:
    global _IS_VERBOSE
    _IS_VERBOSE = False
    _update_log_level()


def is_verbose() -> bool:
    return _IS_VERBOSE


def configure_logging() -> None:
    handler = logging.StreamHandler(sys.stderr)
    formatter = logging.Formatter("[%(levelname)s] %(message)s")
    handler.setFormatter(formatter)

    _update_log_level()
    LOG.addHandler(handler)


def print_exception(e: Exception) -> None:
    if _IS_VERBOSE:
        LOG.exception(e)
    else:
        LOG.error(f"{e!r}")
        LOG.error(VERBOSE_NOTICE)



import os


def is_env_var_true(env_var: str) -> bool:
    val = os.getenv(env_var)
    return val is not None and val.lower() in ("1", "true")


def is_dev_mode_no_margins() -> bool:
    return is_env_var_true("LABELLE_DEV_MODE_NO_MARGINS")


def is_verbose_env_vars() -> bool:
    return is_env_var_true("LABELLE_VERBOSE")



# === LICENSE STATEMENT ===
# Copyright (c) 2011 Sebastian J. Bronner <waschtl@sbronner.com>
#
# Copying and distribution of this file, with or without modification, are
# permitted in any medium without royalty provided the copyright notice and
# this notice are preserved.
# === END LICENSE STATEMENT ===
import contextlib
import logging
import math
import sys
from typing import Generator, List, Tuple

from PIL import ImageDraw

from labelle.lib.constants import PIXELS_PER_MM
from labelle.lib.logger import print_exception

LOG = logging.getLogger(__name__)


def scaling(pix, sc) -> List[Tuple[int, int]]:
    """Scaling pixel up, input: (x,y),scale-factor."""
    return [(pix[0] + i, pix[1] + j) for i in range(sc) for j in range(sc)]


@contextlib.contextmanager
def draw_image(bitmap) -> Generator[ImageDraw.ImageDraw, None, None]:
    drawobj = ImageDraw.Draw(bitmap)
    assert isinstance(drawobj, ImageDraw.ImageDraw)
    try:
        yield drawobj
    finally:
        del drawobj


def px_to_mm(px) -> float:
    mm = px / PIXELS_PER_MM
    # Round up to nearest 0.1mm
    return math.ceil(mm * 10) / 10


def mm_to_px(mm) -> float:
    return mm * PIXELS_PER_MM


@contextlib.contextmanager
def system_run() -> Generator[None, None, None]:
    try:
        yield
    except Exception as e:  # noqa: BLE001
        print_exception(e)
        sys.exit(1)



from __future__ import annotations

import barcode as barcode_module
from PIL import Image

from labelle.lib.barcode_to_image import convert_binary_string_to_barcode_image
from labelle.lib.barcode_writer import SimpleBarcodeWriter
from labelle.lib.constants import DEFAULT_BARCODE_TYPE, BarcodeType
from labelle.lib.render_engines.render_context import RenderContext
from labelle.lib.render_engines.render_engine import RenderEngine, RenderEngineException

if DEFAULT_BARCODE_TYPE != BarcodeType.CODE128:  # pragma: no cover
    # Ensure that we fail fast if the default barcode type is adjusted
    # and the code below hasn't been updated.
    raise RuntimeError(
        "The conditional below assumes that the default barcode type is CODE128. "
        "Different barcodes have different quirks, so we should manually test the "
        "new default to ensure a good user experience in the GUI when the initial "
        "value is an empty string."
    )


class BarcodeRenderError(RenderEngineException):
    def __init__(self, exception: BaseException) -> None:
        msg = f"Barcode render error: {exception!r}"
        super().__init__(msg)


class BarcodeRenderEngine(RenderEngine):
    def __init__(
        self, content: str, barcode_type: BarcodeType = DEFAULT_BARCODE_TYPE
    ) -> None:
        super().__init__()
        self.content = content
        self.barcode_type = barcode_type

    def render(self, context: RenderContext) -> Image.Image:
        if (
            self.barcode_type == DEFAULT_BARCODE_TYPE == BarcodeType.CODE128
            and self.content == ""
        ):
            # An exception is raised on the empty string. Since this is
            # the default code, we really don't want to trigger a popup
            # in the GUI before the user entered a barcode.
            self.content = " "
        try:
            code_obj = barcode_module.get(
                self.barcode_type, self.content, writer=SimpleBarcodeWriter()
            )
            result = code_obj.render()
        except BaseException as e:
            raise BarcodeRenderError(e) from e
        bitmap = convert_binary_string_to_barcode_image(
            line=result.line,
            quiet_zone=result.quiet_zone,
            module_height=context.height_px - 16,
        )
        return bitmap
